/* ==========================================================================
 * Zoned - BIND DLZ Replicator
 * --------------------------------------------------------------------------
 * Copyright (C) 2007  William Ahern
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * ==========================================================================
 */
#include <stddef.h>		/* size_t */

#include <string.h>		/* memcpy(3) */

#include <time.h>		/* time(2) */

#include <pthread.h>		/* PTHREAD_MUTEX_INITIALIZER pthread_mutex_t
				 * */

#include <evnet/thread.h>	/* thread_run() */

#include "log.h"
#include "zmbuf.h"
#include "znode.h"
#include "bdb.h"
#include "edit.h"


#ifndef arraylen
#define arraylen(a)	(sizeof (a) / sizeof (a)[0])
#endif


static struct {
	pthread_mutex_t mutex;
	pthread_cond_t gate;
	SIMPLEQ_HEAD(, zmbuf) queue;
} edits = {
	.mutex	= PTHREAD_MUTEX_INITIALIZER,
	.gate	= PTHREAD_COND_INITIALIZER,
	.queue	= SIMPLEQ_HEAD_INITIALIZER(edits.queue),
};



static size_t edit_rdata_any(unsigned char *buf, size_t bufsiz, struct ZonedRR *rr)  {
	unsigned long repid	= time(0);
	char *host		= (char *)rr->any.host.buf;
	int hlen		= (int)rr->any.host.size;
	unsigned long ttl	= rr->any.ttl;
	char *rtype		= (char *)rr->any.rtype.buf;
	int rtlen		= (int)rr->any.rtype.size;
	char *rdata		= (char *)rr->any.rdata.buf;
	int rdlen		= (int)rr->any.rdata.size;
	int num;

	num	= snprintf((char *)buf, bufsiz, "%lu %.*s %lu %.*s %.*s", repid, hlen, host, ttl, rtlen, rtype, rdlen, rdata);

	return (num < 0)? 0 : num;
} /* edit_rdata_any() */


static size_t edit_rdata_a(unsigned char *buf, size_t bufsiz, struct ZonedRR *rr)  {
	unsigned long repid	= time(0);
	char *host		= (char *)rr->a.host.buf;
	int hlen		= (int)rr->a.host.size;
	unsigned long ttl	= rr->a.ttl;
	char *address		= (char *)rr->a.address.buf;
	int alen		= (int)rr->a.address.size;
	int num;

	num	= snprintf((char *)buf, bufsiz, "%lu %.*s %lu A %.*s", repid, hlen, host, ttl, alen, address);

	return (num < 0)? 0 : num;
} /* edit_rdata_a() */


static size_t edit_rdata_srv(unsigned char *buf, size_t bufsiz, struct ZonedRR *rr)  {
	unsigned long repid	= time(0);
	char *host		= (char *)rr->srv.host.buf;
	int hlen		= (int)rr->srv.host.size;
	unsigned long ttl	= rr->srv.ttl;
	unsigned long priority	= rr->srv.priority;
	unsigned long weight	= rr->srv.weight;
	unsigned short port	= rr->srv.port;
	char *target		= (char *)rr->srv.target.buf;
	int tlen		= (int)rr->srv.target.size;
	int num;

	num	= snprintf((char *)buf, bufsiz, "%lu %.*s %lu SRV %lu %lu %hu %.*s", repid, hlen, host, ttl, priority, weight, port, tlen, target);

	return (num < 0)? 0 : num;
} /* edit_rdata_srv() */



static size_t edit_rdata(void *buf, size_t bufsiz, struct ZonedRR *rr)  {
	static const typeof (&edit_rdata_any) rdata[]	= {
		[ZonedRR_PR_any]	= &edit_rdata_any,
		[ZonedRR_PR_a]		= &edit_rdata_a,
		[ZonedRR_PR_srv]	= &edit_rdata_srv,
	};
	if (rr->present < 0 || rr->present >= arraylen(rdata) || 0 == rdata[rr->present]) {
		error("unknown RR type: %d", rr->present);

		return 0;
	}

	return rdata[rr->present](buf, bufsiz, rr);
} /* edit_rdata() */


static size_t edit_rzone(unsigned char *dstbuf, size_t dstsiz, unsigned char *srcbuf, size_t srcsiz) {
	struct { unsigned char *beg, *pos, *end; } dst, src;

	dst.pos	= dstbuf;
	dst.end	= dstbuf + dstsiz;

	src.pos	= srcbuf + srcsiz;
	src.beg	= srcbuf;

	while (dst.pos < dst.end && src.pos > src.beg) {
		src.pos--;

		*dst.pos	= *src.pos;

		dst.pos++;
	}

	return dst.pos - dstbuf;
} /* edit_rzone() */


static int edit_write(struct zmbuf *mb) {
	unsigned char buf[1024];
	size_t len;
	int i, n, dberr;
	DBT key, data;
	struct ZonedRR *rr;

	data.data	= 0;
	data.size	= 0;

	key.data	= buf;
	key.size	= edit_rzone(buf, sizeof buf, mb->msg->edit.zone.buf, mb->msg->edit.zone.size);

	//debug1("marking zone: %.*s", (int)key.size, (char *)key.data);
	debug1("marking zone: %.*s", (int)mb->msg->edit.zone.size, (char *)mb->msg->edit.zone.buf);
again:
	dberr	= DNSDB.db.zone->put(DNSDB.db.zone, 0, &key, &data, DB_NOOVERWRITE|DB_AUTO_COMMIT);

	if (dberr != 0 && dberr != DB_KEYEXIST) {
		switch (dberr) {
		/* FIXME: Farm this out. */
		case DB_REP_HANDLE_DEAD:
			bdb_openall(&DNSDB, 1);

			goto again;
		default:
			error("unable to mark zone: %s", db_strerror(dberr));
		}
	}

	for (i = 0; i < mb->msg->edit.rrset.list.count; i++) {
		rr	= mb->msg->edit.rrset.list.array[i];

		if (0 >= (n = snprintf((char *)buf, sizeof buf, "%.*s %.*s", (int)mb->msg->edit.zone.size, (char *)mb->msg->edit.zone.buf, (int)rr->any.host.size, rr->any.host.buf))
		||  n >= sizeof buf) {
			warn("RR zone + host too large");

			continue;
		}

		key.data	= buf;
		key.size	= n;

		if (0 == (len = edit_rdata(&buf[n], sizeof buf - n, rr)) || len > sizeof buf - n) {
			warn("unable to write out RR");

			continue;
		}

		data.data	= &buf[n];		
		data.size	= len;

		debug1("storing RR: %.*s", (int)data.size, (char *)data.data);

		/* FIXME: Handle DB_REP_HANDLE_DEAD */
		dberr	= DNSDB.db.data->put(DNSDB.db.data, 0, &key, &data, DB_AUTO_COMMIT);

		if (dberr != 0)
			error("unable to store RR: %s", db_strerror(dberr));
	}

	return 0;
} /* edit_write() */


static void *edit_writer(void *arg0, void *arg1, unsigned arg2) {
	struct zmbuf *mb0, *mb;
	int masterid;

	pthread_mutex_lock(&edits.mutex);

	for (;;) {
		if (0 < (masterid = DNSDB.master.envid) && SIMPLEQ_END(&edits.queue) != (mb0 = SIMPLEQ_FIRST(&edits.queue))) {
			do {
				if (SIMPLEQ_END(&edits.queue) == (mb = SIMPLEQ_FIRST(&edits.queue)))
					break;

				SIMPLEQ_REMOVE_HEAD(&edits.queue, sqe);

				/*
				 * Unlock so others can call edit_apply(). 
				 * Otherwise there may be a deadlock lurking
				 * betwixt edit_apply() and zmbuf_send().
				 */
				pthread_mutex_unlock(&edits.mutex);

				if (masterid == BDB_SELF_EID) {
					if (0 != edit_write(mb)) {
						warn("unable to write out edit");
					} else
						zmbuf_free(mb), mb = 0;
				} else {
					if (0 == zmbuf_send(mb, masterid, 0))
						mb	= 0;
				}

				pthread_mutex_lock(&edits.mutex);

				if (mb != 0)
					SIMPLEQ_INSERT_TAIL(&edits.queue, mb, sqe);
			} while (mb0 != mb);
		}

		if (0 != pthread_cond_wait(&edits.gate, &edits.mutex))
			panic("pthread_cond_wait: %s", strerror(errno));
	}

	pthread_mutex_unlock(&edits.mutex);

	return 0;
} /* edit_writer() */


void edit_apply(struct zmbuf *mb) {
	(void)pthread_mutex_lock(&edits.mutex);

	SIMPLEQ_INSERT_TAIL(&edits.queue, mb, sqe);

	(void)pthread_cond_signal(&edits.gate);

	(void)pthread_mutex_unlock(&edits.mutex);

	return /* void */;
} /* edit_apply() */


void edit_wake(void) {
	(void)pthread_mutex_lock(&edits.mutex);

	(void)pthread_cond_signal(&edits.gate);

	(void)pthread_mutex_unlock(&edits.mutex);

	return /* void */;
} /* edit_wake() */


void edit_init(void) {
	thread_run(&edit_writer, 0, THREAD_PRI_NORM, 0, 0, 0);

	return /* void */;
} /* edit_init() */

