/* ==========================================================================
 * Zoned - BIND DLZ Replicator
 * --------------------------------------------------------------------------
 * Copyright (C) 2007  William Ahern
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * ==========================================================================
 */
#include <stdlib.h>		/* malloc(3) free(3) */

#include <time.h>		/* struct timeval */

#include <sys/types.h>		/* socklen_t */
#include <sys/socket.h>		/* struct sockaddr_storage */

#include <sys/queue.h>		/* LIST */

#include <event.h>

#include <evnet/lookup.h>
#include <evnet/socket.h>

#include "options.h"
#include "log.h"
#include "znode.h"
#include "bdb.h"
#include "locate.h"


#ifndef ZONED_LOCATE_INTERVAL
#define ZONED_LOCATE_INTERVAL	30
#endif


static struct {
	struct { 
		int pending;

		struct event evb;
	} timer;

	int pending;
	int nsites;
} locate;


struct zoned {
	char name[sizeof ((struct lookup_rr *)0)->rr.srv.host];
	unsigned short port;

	struct sockaddr_storage ss;
	socklen_t salen;

	LIST_ENTRY(zoned) le;
}; /* struct zoned */


static void locate_connected(struct socket *s, enum socket_errno s_err, void *arg) {
	struct zoned *zd	= arg;
	struct znode *z;

	if (s_err != 0) {
		notice("unable to connect ([%s]:%hu): %s", zd->name, zd->port, (s_err == SOCKET_ESYSTEM)? strerror(errno) : socket_strerror(s_err));

		socket_close(s);
	} else if (0 == (z = znode_open(s))) {
		error("unable to open znode ([%s]:%hu)", zd->name, zd->port);

		socket_close(s);
	} else
		znode_start(z);

	free(zd);

	return /* void */;
} /* locate_connected() */


static void locate_catch(int nans, struct lookup_rr *ans, int nadd, struct lookup_rr *add, enum lookup_errno lu_err, void *arg) {
	static const struct zoned zd_initializer;
	LIST_HEAD(, zoned) zoneds	= LIST_HEAD_INITIALIZER(&zoneds);
	int nsites			= 0;
	struct zoned *zd		= 0;
	struct lookup_rr *rr;

	locate.pending	= 0;

	for (rr = ans; rr != 0; rr = rr->next) {
		if (rr->type == LOOKUP_IN_SRV) {
			nsites++;

			if (0 != znode_find_byhost(rr->rr.srv.host, rr->rr.srv.port)) {
				zd	= 0;

				continue;
			}

			if (0 == (zd = malloc(sizeof *zd)))
				panic("malloc: %s", strerror(errno));

			*zd	= zd_initializer;

			(void)memcpy(zd->name, rr->rr.srv.host, rr->rr.srv.hostlen);
			zd->port	= rr->rr.srv.port;

			LIST_INSERT_HEAD(&zoneds, zd, le);
		} else if (rr->type == LOOKUP_IN_A && zd != 0) {
			(void)memcpy(&zd->ss, &rr->rr.ip.sa.ss, sizeof zd->ss);
			zd->salen	= rr->rr.ip.salen;

			((struct sockaddr_in *)&zd->ss)->sin_port	= htons(zd->port);
		}
	}

	while (LIST_END(&zoneds) != (zd = LIST_FIRST(&zoneds))) {
		struct timeval timeout		= { 10, 0 };
		struct socket_options s_opts	= socket_defaults;
		struct socket_name s_name;
		enum socket_errno s_err;
		struct socket *s;
		struct znode *z;

		LIST_REMOVE(zd, le);

		if (0 != (z = znode_find_byhost(zd->name, zd->port))) {
			info("already connected ([%s]:%hu)", zd->name, zd->port);

			free(zd);

			continue;
		}

		debug1("contacting ([%s]:%hu)", zd->name, zd->port);

		socket_name_init_addr(&s_name, (struct sockaddr *)&zd->ss, zd->salen);

		if (0 == (s = socket_open(&s_name, &s_opts, 0, 0, &s_err)))
			panic("socket_open: %s", socket_strerror(s_err));

		socket_connect(s, &locate_connected, zd, &timeout);
	}

	locate.nsites	= nsites;

	return /* void */;
} /* locate_catch() */


static void locate_wake(int fd, short events, void *arg) {
	struct timeval it_timeout	= { ZONED_LOCATE_INTERVAL, 0 };
	struct timeval lu_timeout	= { ZONED_LOCATE_INTERVAL / 2, 0 };

	locate.timer.pending	= 0;

	luaZ_getglobal(opts.L, "db", "rep", "in-srv");

	if (lua_isstring(opts.L, -1)) {
		locate.pending	= 1;

		lookup(lua_tostring(opts.L, -1), strlen(lua_tostring(opts.L, -1)), LOOKUP_IN_SRV|LOOKUP_IN_A, 0, &locate_catch, 0, &lu_timeout);
	}

	lua_pop(opts.L, 1);

	timeout_set(&locate.timer.evb, &locate_wake, 0);

	if (0 != timeout_add(&locate.timer.evb, &it_timeout))
		panic("timeout_add: %s", strerror(errno));

	locate.timer.pending	= 1;

	return /* void */;
} /* locate_wake() */


int locate_howmany(void) {
	return locate.nsites;
} /* locate_howmany() */


void locate_init(void) {
	lookup_init(0, 0, 0);

	locate_wake(0, 0, 0);

	return /* void */;
} /* locate_init() */
