/* ==========================================================================
 * log.c - Yet another logging API
 * --------------------------------------------------------------------------
 * Copyright (c) 2007  William Ahern
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ==========================================================================
 */
#include <stdarg.h>	/* va_list va_start va_arg va_end va_copy */
#include <stddef.h>	/* NULL */
#include <stdio.h>	/* stderr fprintf(3) fputc(3) flockfile(3) funlockfile(3) */

#include <string.h>	/* strrchr(3) */

#include <errno.h>	/* errno */

#include <sys/queue.h>
#if !defined LIST_FOREACH
#define LIST_FOREACH(var, head, field) for ((var) = (head)->lh_first; (var) != 0; (var) = (var)->field.le_next)
#endif

#if HAVE_LIBNOSTD
#include <sys/param.h>	/* arraylen */
#elif !defined arraylen
#define arraylen(a)	(sizeof (a) / sizeof (a)[0])
#endif

#include "log.h"


void logger_add_listener(struct logger *l, struct logger_listener *ll) {
	LIST_INSERT_HEAD(&l->listeners, ll, le);

	return /* void */;
} /* logger_add_listener() */


logger_event_t logger_new_typename(struct logger *l, const char *name) {
	int i;

	for (i = 0; i < arraylen(l->typenames); i++) {
		if (0 == l->typenames[i]) {
			l->typenames[i]	= name;

			return 1ULL << i;
		}
	}

	return 0;
} /* logger_new_typename() */


const char *logger_get_typename(struct logger *l, logger_event_t e) {
	const char *name;

	if (0 == (name = l->typenames[LOG_EVENT_BITPOS(LOG_EVENT_GET_PRIMARY(e))]))
		name	= "-";

	return name;
} /* logger_get_typename() */


void logger_vlog(struct logger *l, logger_event_t t, enum logger_priority p, struct logger_aux *a, const char *fmt, va_list ap) {
	struct logger_event e;
	struct logger_listener *ll;
	int errno_;

	e.type		= t;
	e.priority	= p;

	e.aux		= a;

	e.msg.fmt	= fmt;

	LIST_FOREACH(ll, &l->listeners, le) {
		if ((ll->events & t) && (ll->priorities & (1ULL << p))) {
			va_copy(e.msg.ap, ap);

			errno_	= errno;

			ll->cb.record(l, &e, ll->cb.arg);

			errno	= errno_;
		}
	}

	return /* void */;
} /* logger_vlog() */


void logger_log(struct logger *l, logger_event_t t, enum logger_priority p, struct logger_aux *a, const char *fmt, ...) {
	va_list ap;

	va_start(ap, fmt);

	vlog(l, t, p, a, fmt, ap);

	va_end(ap);

	return /* void */;
} /* logger_log() */


static void stdlog_record(struct logger *l, struct logger_event *e, void *arg) {
	struct logger_listener *ll	= arg;

	flockfile(stderr);

	if (stdlog.opts.argv0 != 0)
		(void)fprintf(stderr, "%s:", stdlog.opts.argv0);

	/* For LOGGER_P_DEBUG1 and lower output source information. */
	if (ll->priorities & ~((1ULL << LOGGER_P_DEBUG1) - 1)) {
		const char *file	= "-";
		long line		= 0;

		if (e->aux) {
			if (e->aux->file) {
				if (0 != (file = strrchr(e->aux->file, '/'))) {
					if (*(++file) == '\0')
						file	= "-";
				} else
					file	= e->aux->file;
			}

			line	= e->aux->line;
		}

		(void)fprintf(stderr, "%s:%ld:", file, line);
	}

	(void)fprintf(stderr, "%s: ", logger_typename(l, e->type));

	(void)vfprintf(stderr, e->msg.fmt, e->msg.ap);

	(void)fputc('\n', stderr);

	funlockfile(stderr);

	return /* void */;
} /* stdlog_record() */


static const char stdlog_opts_argv0_initializer[]	= "-";

static void stdlog_record_stub(struct logger *l, struct logger_event *e, void *arg) {
	struct logger_listener *ll	= arg;

	if (stdlog.opts.argv0 == stdlog_opts_argv0_initializer) {
#if _GNU_SOURCE
		stdlog.opts.argv0	= program_invocation_short_name;
#else
		stdlog.opts.argv0	= 0;
#endif
	}

	if (ll->events && ll->priorities) {
		l->listeners.lh_first		= &stdlog.listener;

		stdlog.listener.le.le_next	= ll->le.le_next;
		ll->le.le_next			= &stdlog.listener;
	} else {
		l->listeners.lh_first	= ll->le.le_next;
	}

	return /* void */;
} /* stdlog_record_stub() */


static struct logger_listener stdlog_listener_shim = {
	.events		= ~0ULL,
	.priorities	= ~0ULL,

	.cb	= {
		.arg	= &stdlog_listener_shim,
		.record	= &stdlog_record_stub,
	},
}; /* stdlog_listener_shim */


struct stdlog_logger stdlog = {
	.logger	= {
		.listeners	= {
			&stdlog_listener_shim,
		},

		.typenames	= {
			"fatal",
			"critical",
			"error",
			"warn",
			"notice",
			"info",
			"debug",
		},
	},

	.listener	= {
		.events		= STDLOG_E_DEBUG | (STDLOG_E_DEBUG - 1),
		.priorities	= (1ULL << LOGGER_P_INFO) | ((1ULL << LOGGER_P_INFO) - 1),

		.cb	= {
			.arg	= &stdlog.listener,
			.record	= &stdlog_record,
		},
	},

	.opts		= {
		.argv0	= stdlog_opts_argv0_initializer,
	},

	.e	= {
		.fatal		= STDLOG_E_FATAL,
		.critical	= STDLOG_E_CRITICAL,
		.error		= STDLOG_E_ERROR,
		.warn		= STDLOG_E_WARN,
		.notice		= STDLOG_E_NOTICE,
		.info		= STDLOG_E_INFO,
		.debug		= STDLOG_E_DEBUG,
	},
}; /* stdlog */

