/* ==========================================================================
 * log.h - Yet another logging API
 * --------------------------------------------------------------------------
 * Copyright (c) 2007  William Ahern
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ==========================================================================
 */
#ifndef LOG_H

#include <stdarg.h>		/* va_list */

#include <stdlib.h>		/* EXIT_FAILURE _Exit(3) */

#include <string.h>		/* strerror(3) */

#include <errno.h>		/* errno */

#include <sys/queue.h>

#if HAVE_LIBNOSTD
#include <sys/param.h>		/* PP_NARG */
#elif !defined PP_NARG
/*
 * (Laurent Deniau, news:comp.std.c, Jan 16, 2006)
 */
#define PP_NARG(...)	PP_NARG_(__VA_ARGS__,PP_RSEQ_N())
#define PP_NARG_(...)	PP_ARG_N(__VA_ARGS__)

#define PP_ARG_N( \
	_1, _2, _3, _4, _5, _6, _7, _8, _9,_10, \
	_11,_12,_13,_14,_15,_16,_17,_18,_19,_20, \
	_21,_22,_23,_24,_25,_26,_27,_28,_29,_30, \
	_31,_32,_33,_34,_35,_36,_37,_38,_39,_40, \
	_41,_42,_43,_44,_45,_46,_47,_48,_49,_50, \
	_51,_52,_53,_54,_55,_56,_57,_58,_59,_60, \
	_61,_62,_63,N,...) N

#define PP_RSEQ_N() \
	63,62,61,60,                   \
	59,58,57,56,55,54,53,52,51,50, \
	49,48,47,46,45,44,43,42,41,40, \
	39,38,37,36,35,34,33,32,31,30, \
	29,28,27,26,25,24,23,22,21,20, \
	19,18,17,16,15,14,13,12,11,10, \
	9,8,7,6,5,4,3,2,1,0 
#endif


#define LOGGER_VERSION		2007041401UL


#define LOGGER_MAX_TYPENAMES	((1ULL << LOG_EVENT_MLEN) - LOG_EVENT_MLEN)


/*
 * Composition for multiple events. Maximum of 8 supported.
 */
#define LOG_EVENT_SET_PRIMARY(e)	((1 + (unsigned long long)LOG_EVENT_BITPOS(e)) << ((1ULL << LOG_EVENT_MLEN) - LOG_EVENT_MLEN))
#define LOG_EVENT_GET_PRIMARY(e)						\
	(((e) >> ((1ULL << LOG_EVENT_MLEN) - LOG_EVENT_MLEN))			\
	 ? (1ULL << (((e) >> ((1ULL << LOG_EVENT_MLEN) - LOG_EVENT_MLEN)) - 1))	\
	 : ((e) & -(e)))

#define LOG_EVENTS_N(e0, e1, e2, e3, e4, e5, e6, e7, ...)	\
	(e0) | (e1) | (e2) | (e3) | (e4) | (e5) | (e6) | (e7)	\
	| LOG_EVENT_SET_PRIMARY(e0)

#define LOG_EVENTS(...)	LOG_EVENTS_N(__VA_ARGS__, 0, 0, 0, 0, 0, 0, 0)
#define E(...)		LOG_EVENTS(__VA_ARGS__)

/*
 * Magic hash value used to map power-of-2 values to their bit positions.
 *
 * http://groups.google.com/group/comp.lang.c/msg/0cef41f343f0e875
 *
 * The hash and map herein is sufficient up to 64-bits, derived
 * using the following purpose-built generator:
 *
 * static void overlay(unsigned long m, unsigned long depth, unsigned long have) {
 * 	unsigned long i, bit;
 * 
 * 	if (prove(m))
 * 		{ print_results(m); exit(0); }
 * 
 * 	for (i = 1, bit = 1; i < NUMBITS; i++, bit <<= 1) {
 * 		if (have & bit)
 * 			continue;
 * 
 * 		if (i != ((0x01UL & i) | (BITMASK & m)))
 * 			continue;
 * 
 * 		overlay((m | i) << 1, depth + 1, have | bit);
 * 	}
 * 
 * 	return;
 * }
 *
 * void main(void) { overlay(0, 0, 0); }
 */
#define LOG_EVENT_HASH	0x43147259a7abb7eULL	/* Magic number. */
#define LOG_EVENT_MLEN	6			/* Mask length. */

#define LOG_EVENT_BITIDX(e)	(((1ULL << LOG_EVENT_MLEN) - 1) & ((((e) & -(e)) * LOG_EVENT_HASH) >> ((1ULL << LOG_EVENT_MLEN) - LOG_EVENT_MLEN)))
#define LOG_EVENT_BITPOS(e)	logger_event_bitpos[LOG_EVENT_BITIDX(e)]

static const int logger_event_bitpos[] = {
	63,  0,  1,  6,  2, 12,  7, 18,  3, 24, 13, 27,  8, 33, 19, 39,
	 4, 16, 25, 37, 14, 45, 28, 47,  9, 30, 34, 53, 20, 49, 40, 56,
	62,  5, 11, 17, 23, 26, 32, 38, 15, 36, 44, 46, 29, 52, 48, 55,
	61, 10, 22, 31, 35, 43, 51, 54, 60, 21, 42, 50, 59, 41, 58, 57,
}; /* logger_event_bitpos[] */



typedef unsigned long long logger_event_t;


enum logger_priority {
	LOGGER_P_FATAL		= 0,
	LOGGER_P_CRITICAL	= 1,
	LOGGER_P_ERROR		= 5,
	LOGGER_P_WARN		= 10,
	LOGGER_P_NOTICE		= 15,
	LOGGER_P_INFO		= 20,
	LOGGER_P_DEBUG		= 25,
	LOGGER_P_DEBUG0		= 25,
	LOGGER_P_DEBUG1,
	LOGGER_P_DEBUG2,

	LOGGER_P_LOWEST		= 31,
}; /* enum logger_priority */


struct logger_aux {
	const char *file;
	long line;

	void *aux[];
}; /* struct logger_aux */


struct logger_event {
	logger_event_t type;

	enum logger_priority priority;

	struct {
		const char *fmt;
		va_list ap;
	} msg;

	struct logger_aux *aux;
}; /* struct logger_event */


struct logger {
	LIST_HEAD(, logger_listener) listeners;

	const char *typenames[LOGGER_MAX_TYPENAMES];
}; /* struct logger */


struct logger_listener {
	logger_event_t events;

	unsigned long long priorities;

	struct {
		void *arg;

		void (*record)(struct logger *, struct logger_event *, void *);
	} cb;

	LIST_ENTRY(logger_listener) le;
}; /* struct logger_listener */


void logger_add_listener(struct logger *, struct logger_listener *);

logger_event_t logger_new_typename(struct logger *, const char *);

void logger_set_typename(struct logger *, logger_event_t, const char *);

const char *logger_get_typename(struct logger *, logger_event_t);

#define logger_typename(l, e)	logger_get_typename((l), (e))


#ifndef log
#define log(...)	logger_log(__VA_ARGS__)
#endif

void logger_log(struct logger *, logger_event_t, enum logger_priority, struct logger_aux *, const char *fmt, ...)
	__attribute__((__format__ (printf, 5, 6)));

#ifndef vlog
#define vlog(...)	logger_vlog(__VA_ARGS__)
#endif

void logger_vlog(struct logger *, logger_event_t, enum logger_priority, struct logger_aux *, const char *fmt, va_list)
	__attribute__((__format__ (printf, 5, 0)));


#define LOG_AUX_SRC_ARGS	__FILE__, __LINE__
#define LOG_AUX_STD_ARGS	LOG_AUX_SRC_ARGS
#define LOG_AUX_STD		log_aux(LOG_AUX_STD_ARGS)

#if 0
#include <alloca.h>
#define log_aux_new(n)	((struct logger_aux *)alloca(sizeof (void *) * (n)))
#else
#define log_aux_new(n)	((struct logger_aux *)((void *[(n)]){ 0 }))
#endif

#define log_aux_init(a, ...)	log_aux_initN(a, PP_NARG(__VA_ARGS__), __VA_ARGS__)

#define log_aux(...)		log_aux_init(log_aux_new(PP_NARG(__VA_ARGS__)), __VA_ARGS__)

#define log_stdaux(...)		log_aux(LOG_AUX_STD_ARGS, __VA_ARGS__)

static inline struct logger_aux *log_aux_initN(struct logger_aux *a, unsigned n, const char *file, long line, ...) {
	va_list ap;
	void **aux;

	a->file	= file;
	a->line	= line;

	if (n == 2)
		return a;

	va_start(ap, line);

	for (aux = a->aux; n > 2; n--, aux++)
		*aux	= va_arg(ap, typeof(*aux));

	va_end(ap);

	return a;
} /* log_aux_initN() */


/*
 * Use literals for the most common events, so our complex event type bit
 * fiddling scheme can be constant folded as much as possible.
 */
#define STDLOG_E_FATAL		(1ULL << 0)
#define STDLOG_E_CRITICAL	(1ULL << 1)
#define STDLOG_E_ERROR		(1ULL << 2)
#define STDLOG_E_WARN		(1ULL << 3)
#define STDLOG_E_NOTICE		(1ULL << 4)
#define STDLOG_E_INFO		(1ULL << 5)
#define STDLOG_E_DEBUG		(1ULL << 6)


extern struct stdlog_logger {
	struct logger logger;

	struct logger_listener listener;

	struct {
		const char *argv0;
	} opts;

	const struct {
		logger_event_t fatal;
		logger_event_t critical;
		logger_event_t error;
		logger_event_t warn;
		logger_event_t notice;
		logger_event_t info;
		logger_event_t debug;
	} e;
} stdlog;


#define stdlog(...)	log(&stdlog.logger, __VA_ARGS__)
#define stdvlog(...)	vlog(&stdlog.logger, __VA_ARGS__)


#define fatal(...)	stdlog(STDLOG_E_FATAL, LOGGER_P_FATAL, LOG_AUX_STD, __VA_ARGS__)
#define critical(...)	stdlog(STDLOG_E_CRITICAL, LOGGER_P_CRITICAL, LOG_AUX_STD, __VA_ARGS__)
#define error(...)	stdlog(STDLOG_E_ERROR, LOGGER_P_ERROR, LOG_AUX_STD, __VA_ARGS__)
#define warn(...)	stdlog(STDLOG_E_WARN, LOGGER_P_WARN, LOG_AUX_STD, __VA_ARGS__)
#define notice(...)	stdlog(STDLOG_E_NOTICE, LOGGER_P_NOTICE, LOG_AUX_STD, __VA_ARGS__)
#define info(...)	stdlog(STDLOG_E_INFO, LOGGER_P_INFO, LOG_AUX_STD, __VA_ARGS__)
#define debug(...)	stdlog(STDLOG_E_DEBUG, LOGGER_P_DEBUG, LOG_AUX_STD, __VA_ARGS__)
#define debug0(...)	stdlog(STDLOG_E_DEBUG, LOGGER_P_DEBUG + 0, LOG_AUX_STD, __VA_ARGS__)
#define debug1(...)	stdlog(STDLOG_E_DEBUG, LOGGER_P_DEBUG + 1, LOG_AUX_STD, __VA_ARGS__)
#define debug2(...)	stdlog(STDLOG_E_DEBUG, LOGGER_P_DEBUG + 2, LOG_AUX_STD, __VA_ARGS__)

#define panic(...)	do { fatal(__VA_ARGS__); _Exit(EXIT_FAILURE); } while (0);

#endif /* LOG_H */
