/* ==========================================================================
 * Zoned - BIND DLZ Replicator
 * --------------------------------------------------------------------------
 * Copyright (C) 2007  William Ahern
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * ==========================================================================
 */
#include <stdlib.h>	/* daemon(3) setenv(3) atexit(3) exit(3) */

#include <stdio.h>	/* freopen(3) fopen(3) popen(3) fileno(3) */

#include <time.h>	/* tzset(3) */

#include <string.h>	/* strcmp(3) string(3) strsignal(3) */

#include <errno.h>	/* errno */

#include <signal.h>	/* SIGPIPE SIGHUP SIGTERM SIGKILL SIG_IGN struct sigaction sigemptyset(3) sigaction(2) kill(2) */

#include <sys/param.h>	/* MIN */

#include <sys/types.h>	/* pid_t */

#include <sys/wait.h>	/* WEXITSTATUS WIFSIGNALLED WTERMSIG waitpid(2) */

#include <unistd.h>	/* STDERR_FILENO getpid(2) setuid(2) setgid(2) dup2(2) fork(2) sleep(3) */

#include <event.h>	/* event_init() */

#include <evnet/thread.h>

#include "options.h"
#include "pidfile.h"
#include "listen.h"
#include "znode.h"
#include "locate.h"
#include "bdb.h"
#include "edit.h"
#include "log.h"


static struct pidfile *pf;

static FILE *efp;


static void pidfile_exit(void) {
	if (pf != 0)
		pidfile_close(pf, 1), pf = 0;

	return /* void */;
} /* pidfile_exit() */


static void watchdog_sighandler(int signum) {
	if (signum != SIGCHLD)
		exit(EXIT_SUCCESS);

	return /* void */;
} /* watchdog_sighandler() */


static void watched_exit(int fd, short events, void *arg) {
	exit(EXIT_SUCCESS);

	return /* void */;
} /* watched_exit() */


static void watchdog_start(void) {
	static const struct sigaction sa_initializer;
	static struct event ev;
	struct sigaction sa;
	int fd[2];
	pid_t pid;
	int status;
	int nforked	= 0;

	sa		= sa_initializer;
	sa.sa_handler	= &watchdog_sighandler;
	sa.sa_flags	= 0;

	sigemptyset(&sa.sa_mask);

	if (0 != sigaction(SIGHUP, &sa, 0))
		panic("sigaction: %s", strerror(errno));

	if (0 != sigaction(SIGTERM, &sa, 0))
		panic("sigaction: %s", strerror(errno));

	if (0 != sigaction(SIGCHLD, &sa, 0))
		panic("sigaction: %s", strerror(errno));

	do {
		if (0 != pipe(fd))
			panic("pipe: %s", strerror(errno));

		switch ((pid = fork())) {
		case -1:
			panic("fork: %s", strerror(errno));
		case 0:
			close(fd[1]), fd[1] = -1;

			if (pf != 0)
				pidfile_close(pf, 0), pf =  0;

			event_set(&ev, fd[0], EV_READ, &watched_exit, 0);

			if (0 != event_add(&ev, 0))
				panic("event_add: %s", strerror(errno));

			return /* void */;
		default:
			break;
		}

		close(fd[0]), fd[0] = -1;

		notice("forked process: %d", (int)pid);

		nforked++;

		do {
			if (pid != waitpid(pid, &status, 0)) {
				if (errno != EINTR)
					panic("waitpid: %s", strerror(errno));
			} else {
				if (WIFSIGNALED(status)) {
					critical("process %d terminated by signal %d: %s", (int)pid, (int)WTERMSIG(status), strsignal(WTERMSIG(status)));
				} else if (0 != WEXITSTATUS(status)) {
					critical("process %d exited with abnormal status: %d", (int)pid, (int)WEXITSTATUS(status));
				} else
					notice("process %d exited cleanly", (int)pid);

				pid	= -1;
			}
		} while (pid > 0);

		close(fd[1]), fd[1] = -1;

		sleep(MIN(5, nforked));
	} while (1);

	return /* void */;
} /* watchdog_start() */


static void stderr_open(const char *to) {
	if (*to == '|') {
		to++;

		if (0 == (efp = popen(to, "w")))
			panic("popen(`%s'): %s", to, strerror(errno));
	} else {
		if (0 == (efp = fopen(to, "a")))
			panic("fopen(`%s'): %s", to, strerror(errno));
	}

	if (0 > dup2(fileno(efp), STDERR_FILENO))
		panic("dup2(%d, %d): %s", fileno(efp), STDERR_FILENO, strerror(errno));

	return /* void */;
} /* stderr_open() */


static void sig_init(void) {
	struct sigaction sigpipe;

	/*
	 * We are NOT a command line utility.
	 */
	sigpipe.sa_handler	= SIG_IGN;
	sigpipe.sa_flags	= 0;

	sigemptyset(&sigpipe.sa_mask);

	if (0 != sigaction(SIGPIPE, &sigpipe, 0))
		panic("sigaction: %s", strerror(errno));

	return /* void */;
} /* sig_init() */


int main(int argc, char *argv[]) {

	/* Timezones and Unix don't mix. */
	if (0 != setenv("TZ", "UTC", 1))
		panic("setenv(TZ, UTC, 1): %s", strerror(errno));

	tzset();


	/* Parse and sanity check command line options. */
	options_init(argc, argv);


	/* Setup logging as requested */
	luaZ_getglobal(opts.L, "log", "level");

	if (!lua_isnil(opts.L, -1)) {
		unsigned long level	= 1ULL << MIN(31, MAX(0, lua_tointeger(opts.L, -1)));

		stdlog.listener.priorities	= level | (level - 1);
	}

	lua_pop(opts.L, 1);


	/*
	 * Redirect stderr, but after we've setup stdin and stdout to
	 * guarantee the FILE objects match the descriptors (0, 1, 2).
	 */
	if (stdin != freopen("/dev/null", "r", stdin))
		panic("freopen(/dev/null, r, stdin): %s", strerror(errno));

	if (stdout != freopen("/dev/null", "r+", stdout))
		panic("freopen(/dev/null, w, stdout): %s", strerror(errno));

	luaZ_getglobal(opts.L, "log", "stderr");

	if (!lua_isnil(opts.L, -1))
		stderr_open(luaL_optstring(opts.L, -1, ""));

	lua_pop(opts.L, 1);


	/*
	 * Try to open our PID file early
	 */
	luaZ_getglobal(opts.L, "proc", "pid-file");

	if (!lua_isnil(opts.L, -1)) {
		if (0 == (pf = pidfile_open(luaL_optstring(opts.L, -1, ""), &pidfile_defaults)))
			panic("unable to open PID file (`%s'): %s", lua_tostring(opts.L, -1), strerror(errno));
	}

	lua_pop(opts.L, 1);


	/*
	 * Try to acquire the lock sooner rather than later.
	 */
	if (pf != 0) {
		pid_t p	= pidfile_getpid(pf);

		if (p != PIDFILE_VOID_PID)
			debug("PID file purportedly owned by %d", (int)p);

		luaZ_getglobal(opts.L, "main", "do");

		if (0 == strcmp("stop", luaL_optstring(opts.L, -1, ""))) {
			if (0 != pidfile_lock(pf))
				panic("unable to acquire PID file lock: %s", strerror(errno));

			exit(EXIT_SUCCESS);
		} else if (0 == strcmp("restart", luaL_optstring(opts.L, -1, ""))) {
			if (0 != pidfile_lock(pf))
				panic("unable to acquire PID file lock: %s", strerror(errno));
		} else if (0 != pidfile_trylock(pf))
			panic("unable to acquire PID file lock: %s", strerror(errno));

		lua_pop(opts.L, 1);

		if (p != PIDFILE_VOID_PID)
			notice("acquired PID file from %d", (int)p);
	}


	/*
	 * Daemonize
	 */
	luaZ_getglobal(opts.L, "proc", "daemon");

	if (lua_toboolean(opts.L, -1) && 0 != daemon(0, 1))
		panic("daemon(0, 1): %s", strerror(errno));

	lua_pop(opts.L, 1);


	/*
	 * Set our PID
	 */
	if (pf != 0) {
		if (0 != pidfile_setpid(pf, getpid()))
			panic("pidfile_setpid(%d): %s", (int)getpid(), strerror(errno));

		atexit(&pidfile_exit);
	}


	/*
	 * Signal housekeeping
	 */
	sig_init();


	/*
	 * Some library initializations.
	 */
	event_init();


	/*
	 * Watchdog
	 */
	luaZ_getglobal(opts.L, "proc", "watchdog");

	if (lua_toboolean(opts.L, -1))
		watchdog_start();

	lua_pop(opts.L, 1);


	/*
	 * Thread stuff (after we fork)
	 */
	struct thread_options thread_opts	= thread_defaults;

	thread_opts.nthreads_fixed	= 1;

	thread_opts.sigmask_how		= SIG_BLOCK;
	(void)sigfillset(&thread_opts.sigmask_set);

	if (0 != thread_init(&thread_opts, 0))
		panic("thread_init: %s", strerror(errno));


	/*
	 * Edit worker thread
	 */
	edit_init();


	/*
	 * Zoned node backbone.
	 */
	znode_init();

	event_loop(EVLOOP_NONBLOCK);	/* make libevent setup signal handlers */


	/*
	 * Setup our communications channels.
	 */
	listen_init();


	/*
	 * Drop privileges.
	 */
	luaZ_getglobal(opts.L, "proc", "gid");

	if (!lua_isnil(opts.L, -1) && 0 != setgid(lua_tointeger(opts.L, -1)))
		panic("setgid(%d): %s", (int)lua_tointeger(opts.L, -1), strerror(errno));

	luaZ_getglobal(opts.L, "proc", "uid");

	if (!lua_isnil(opts.L, -1) && 0 != setuid(lua_tointeger(opts.L, -1)))
		panic("setuid(%d): %s", (int)lua_tointeger(opts.L, -1), strerror(errno));

	lua_pop(opts.L, 2);


	/*
	 * DB environment.
	 */
	bdb_init();


	/*
	 * Node locator.
	 */
	locate_init();


	/*
	 * Loop....
	 */
	event_dispatch();

	return 0;
} /* main() */
