/* ==========================================================================
 * Zoned - BIND DLZ Replicator
 * --------------------------------------------------------------------------
 * Copyright (C) 2007  William Ahern
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * ==========================================================================
 */
#include <stdarg.h>	/* va_list va_start va_end */

#include <stdlib.h>	/* EXIT_FAILURE EXIT_SUCCESS exit(3) */

#include <stdio.h>	/* FILE fprintf(3) fflush(3) */

#include <limits.h>	/* CHAR_MAX CHAR_MIN */

#include <ctype.h>	/* isalpha */

#include <signal.h>	/* SIGSEGV raise(2) */

#include <string.h>	/* strcmp(3) strchr(3) memcpy(3) */

#include <sys/types.h>	/* uid_t gid_t */
#include <sys/socket.h>	/* struct sockaddr */

#include <unistd.h>	/* geteuid(2) getegid(2) setuid(2) setgid(2) gethostname(3) */

#include <netdb.h>	/* struct addrinfo getaddrinfo(3) */

#include <pwd.h>	/* struct passwd getpwnam(3) */
#include <grp.h>	/* struct group getgrnam(3) */

#include <getopt.h>	/* struct option optarg optval getopt_long(3) */

#include "version.h"
#include "options.h"
#include "log.h"


struct options opts;



static int luaZ_proc__newindex(lua_State *L) {
	if (!lua_isstring(L, 2) || lua_isnumber(L, 3) || lua_isnil(L, 3)) {
rawset:
		/*
		 * Store in proc.__metatable.__index
		 */
		lua_getmetatable(L, 1);
			lua_getfield(L, -1, "__index");
			lua_replace(L, 1);
		lua_pop(L, 1);

		lua_rawset(L, 1);
	} else if (0 == strcmp(lua_tostring(L, 2), "uid")) {
		struct passwd *pw;

		if (0 == (pw = getpwnam(luaL_optstring(L, 3, ""))))
			panic("getpwnam(`%s'): %s", luaL_optstring(L, 3, ""), (errno)? strerror(errno) : "No such user");

		lua_pushinteger(L, pw->pw_uid);
		lua_replace(L, 3);

		goto rawset;
	} else if (0 == strcmp(lua_tostring(L, 2), "gid")) {
		struct group *gr;

		if (0 == (gr = getgrnam(luaL_optstring(L, 3, ""))))
			panic("getgrnam(`%s'): %s", luaL_optstring(L, 3, ""), (errno)? strerror(errno) : "No such group");

		lua_pushinteger(L, gr->gr_gid);
		lua_replace(L, 3);

		goto rawset;
	} else
		goto rawset;

	return 0;
} /* luaZ_proc__newindex() */


static int luaZ_proc_listen__newindex(lua_State *L) {
	struct { char buf[128]; char *pos, *end; } atoms[4], *atom, *host, *port;
	const char *src;
	int quoted	= 0;
	int escaped	= 0;
	int ch, e;
	struct addrinfo *res, *res0, hints	= { .ai_family = AF_UNSPEC, .ai_socktype = SOCK_DGRAM };

	luaL_argcheck(L, (src = lua_tostring(L, 3)), 1, "bind specifier must be string");

	atom	= atoms;

	while (*src != '\0' && atom < &atoms[sizeof atoms / sizeof atoms[0]]) {
		atom->pos	= atom->buf;
		atom->end	= &atom->buf[sizeof atom->buf - 1];

		while (*src != '\0' && atom->pos < atom->end) {
			switch ((ch = *(src++))) {
			case '[':
				if (quoted || escaped)
					goto copy;

				quoted	= '[';

				break;
			case ']':
				if (!quoted || escaped)
					goto copy;

				quoted	= 0;

				break;
			case '\\':
				if (escaped)
					goto copy;

				escaped	= 1;

				break;
			case ':':
				if (quoted || escaped)
					goto copy;

				goto next;
copy:			default:
				escaped	= 0;

				*(atom->pos++)	= ch;
			}
		} /* while() */
next:
		*(atom++)->pos	= '\0';
	} /* for() */

	lua_pop(L, 2);

	switch (atom - atoms) {
	case 1:
		if (0 != gethostname((host = &atoms[1])->buf, sizeof atoms[1].buf))
			panic("gethostname: %s", strerror(errno));

		port	= &atoms[0];

		break;
	case 2:
		host	= &atoms[0];
		port	= &atoms[1];

		break;
	default:
		luaL_argcheck(L, atom - atoms == 2, 1, "address must specify both host and port");
	}

	if (0 != (e = getaddrinfo(host->buf, port->buf, &hints, &res0)))
		panic("getaddrinfo(`%s', `%s'): %s", host->buf, port->buf, gai_strerror(e));

	for (res = res0; 0 != res; res = res->ai_next) {
		luaZ_getglobal(L, "table", "insert");

		lua_pushvalue(L, 1);

		(void)memcpy(lua_newuserdata(L, res->ai_addrlen), res->ai_addr, res->ai_addrlen);

		luaL_newmetatable(L, "sockaddr");
		lua_setmetatable(L, -2);

		lua_call(L, 2, 0);
	}

	freeaddrinfo(res0), res0 = 0;

	return 0;
} /* luaZ_proc_listen__newindex() */


static int luaZ_log__newindex(lua_State *L) {
	if (!lua_isstring(L, 2) || lua_isnumber(L, 3) || lua_isnil(L, 3)) {
rawset:
		/*
		 * Store in log.__metatable.__index
		 */
		lua_getmetatable(L, 1);
			lua_getfield(L, -1, "__index");
			lua_replace(L, 1);
		lua_pop(L, 1);

		lua_rawset(L, 1);
	} else if (0 == strcmp(lua_tostring(L, 2), "level")) {
		static const char *lvl[] = {
			"fatal", "critical", "error",
			"warn", "notice", "info",
			"debug", "debug1", "debug2", 0
		};
		static const int map[] = { 0, 1, 5, 10, 15, 20, 25, 26, 27 };

		lua_pushinteger(opts.L, map[luaL_checkoption(opts.L, 3, 0, lvl)]);
		lua_replace(opts.L, 3);

		goto rawset;
	} else
		goto rawset;

	return 0;
} /* luaZ_log__newindex() */


static int luaZ_panic(lua_State *L) {
#if 1
	panic("unprotected error in call to Lua API (%s)", lua_tostring(L, -1));
#else
	fatal("unprotected error in call to Lua API (%s)", lua_tostring(L, -1));

	raise(SIGSEGV);
#endif

	return 0;
} /* luaZ_panic() */


static void luaZ_init(void) {
	if (0 == (opts.L = luaL_newstate()))
		panic("luaL_newstate: %s", strerror(errno));

	lua_atpanic(opts.L, &luaZ_panic);

	luaL_openlibs(opts.L);

	/*
	 * main
	 */
	lua_newtable(opts.L);

		lua_pushvalue(opts.L, -1);
		lua_setglobal(opts.L, "main");

		/*
		 * main.do
		 */
		lua_pushstring(opts.L, "start");
		lua_setfield(opts.L, -2, "do");

	lua_pop(opts.L, 1);

	/*
	 * proc
	 */
	lua_newtable(opts.L);

		lua_pushvalue(opts.L, -1);
		lua_setglobal(opts.L, "proc");

		/*
		 * proc.__metatable
		 */
		lua_newtable(opts.L);

			lua_pushvalue(opts.L, -1);
			lua_setmetatable(opts.L, -3);

			/*
			 * proc.__metatable.__newindex
			 */
			lua_pushcfunction(opts.L, &luaZ_proc__newindex);
			lua_setfield(opts.L, -2, "__newindex");

			/*
			 * proc.__metatable.__index
			 */
			lua_newtable(opts.L);
			lua_setfield(opts.L, -2, "__index");

		lua_pop(opts.L, 1);

		/*
		 * proc.daemon
		 */
		lua_pushboolean(opts.L, 0);
		lua_setfield(opts.L, -2, "daemon");

		/*
		 * proc.pid-file
		 */
		lua_pushnil(opts.L);
		lua_setfield(opts.L, -2, "pid-file");

		/*
		 * proc.watchdog
		 */
		lua_pushboolean(opts.L, 0);
		lua_setfield(opts.L, -2, "watchdog");

		/*
		 * proc.watchdog
		 */
		lua_pushboolean(opts.L, 0);
		lua_setfield(opts.L, -2, "watchdog");

		/*
		 * proc.uid, proc.gid
		 */
		lua_pushinteger(opts.L, geteuid());
		lua_setfield(opts.L, -2, "uid");

		lua_pushinteger(opts.L, getegid());
		lua_setfield(opts.L, -2, "gid");

		/*
		 * proc.listen
		 */
		lua_newtable(opts.L);
			lua_pushvalue(opts.L, -1);
			lua_setfield(opts.L, -3, "listen");

			/*
			 * proc.listen.__metatable
			 */
			lua_newtable(opts.L);
				lua_pushvalue(opts.L, -1);
				lua_setmetatable(opts.L, -3);

				/*
				 * proc.listen.__metatable.__newindex
				 */
				lua_pushcfunction(opts.L, &luaZ_proc_listen__newindex);
				lua_setfield(opts.L, -2, "__newindex");

			lua_pop(opts.L, 1);

		lua_pop(opts.L, 1);

	lua_pop(opts.L, 1);

	/*
	 * log
	 */
	lua_newtable(opts.L);

		lua_pushvalue(opts.L, -1);
		lua_setglobal(opts.L, "log");

		/*
		 * log.__metatable
		 */
		lua_newtable(opts.L);

			lua_pushvalue(opts.L, -1);
			lua_setmetatable(opts.L, -3);

			/*
			 * log.__metatable.__newindex
			 */
			lua_pushcfunction(opts.L, &luaZ_log__newindex);
			lua_setfield(opts.L, -2, "__newindex");

			/*
			 * log.__metatable.__index
			 */
			lua_newtable(opts.L);
			lua_setfield(opts.L, -2, "__index");

		lua_pop(opts.L, 1);

		/*
		 * log.stderr
		 */
		lua_pushnil(opts.L);
		lua_setfield(opts.L, -2, "stderr");

		/*
		 * log.level
		 */
		lua_pushnil(opts.L);
		lua_setfield(opts.L, -2, "level");

	lua_pop(opts.L, 1);

	/*
	 * db
	 */
	lua_newtable(opts.L);

		lua_pushvalue(opts.L, -1);
		lua_setglobal(opts.L, "db");

		/*
		 * db.env
		 */
		lua_newtable(opts.L);
			lua_pushvalue(opts.L, -1);
			lua_setfield(opts.L, -3, "env");

			/* db.env.dir */
			lua_pushnil(opts.L);
			lua_setfield(opts.L, -2, "dir");

		lua_pop(opts.L, 1);

		/*
		 * db.rep
		 */
		lua_newtable(opts.L);
			lua_pushvalue(opts.L, -1);
			lua_setfield(opts.L, -3, "rep");

			/* db.rep.in-srv */
			lua_pushnil(opts.L);
			lua_setfield(opts.L, -2, "in-srv");

		lua_pop(opts.L, 1);

	lua_pop(opts.L, 1);

	return /* void */;
} /* luaZ_init() */


static void print_version(FILE *fp) {
	(void)fprintf(fp, "zoned %s\n", version_pretty());

	(void)fflush(fp);

	return /* void */;
} /* print_version() */


static void print_usage(FILE *fp) {
	(void)fprintf(fp,
		"Usage: zoned [-cdweulqvVh] start|stop|restart\n"
		"  -c, --config=FILE         Path to configuration file\n"
		"  -d, --daemon              Daemonize\n"
		"  -p, --pid-file=PATH       Path to PID file\n"
		"  -w, --watchdog            Enable fail-safe mode\n"
		"  -e, --stderr              Path or command to direct stderr\n"
		"  -u, --user=USER[.GROUP]   setuid and setgid\n"
		"  -l, --listen=ADDR[:PORT]  Address to bind to\n"
		"  -q, --query=DNSSRV        SRV query which enumerates peers\n"
		"  -D, --data=PATH           Path to DB ENV data directory\n"
		"  -v, --verbose=LEVEL       Adjust/specify verboseness\n"
		"  -o, --option=CODE         Set key=value, or abuse Lua VM\n"
		"  -V, --version             Print version\n"
		"  -h, --help                Print help\n"
		"\n"
		"Report bugs to William Ahern <william@25thandClement.com>\n");

	(void)fflush(fp);

	return /* void */;
} /* print_usage() */


static const char *short_opts(struct option *long_opts) {
	static char opts[128];
	struct option *long_opt;
	char *opt, *end;

	long_opt	= long_opts;

	opt		= opts;
	end		= opts + sizeof opts - 3;

	*(opt++)	= ':';

	while (long_opt->name != 0 && opt < end) {
		if (isalpha(long_opt->val)) {
			*(opt++)	= long_opt->val;

			switch (long_opt->has_arg) {
			case optional_argument:
				*(opt++)	= ':';
			case required_argument:
				*(opt++)	= ':';
			}
		}

		long_opt++;	
	}

	return opts;
} /* short_opts() */


void options_init(int argc, char *argv[]) {
	static struct option long_opts[] = {
		{ "config",	required_argument,	0,	 'c' },
		{ "daemon",	no_argument,		0,	 'd' },
		{ "pid-file",	required_argument,	0,	 'p' },
		{ "watchdog",	no_argument,		0,	 'w' },
		{ "stderr",	required_argument,	0,	 'e' },
		{ "user",	required_argument,	0,	 'u' },
		{ "listen",	required_argument,	0,	 'l' },
		{ "query",	required_argument,	0,	 'q' },
		{ "data",	required_argument,	0,	 'D' },
		{ "",		no_argument,		0,	 'v' },
		{ "verbose",	required_argument,	0,	-'v' },
		{ "option",	required_argument,	0,	 'o' },
		{ "version",	no_argument,		0,	 'V' },
		{ "help",	no_argument,		0,	 'h' },
		{ 0,		0,			0,	'\0' }
	}; /* long_opts[] */

	extern char *optarg;
	extern int optopt, opterr;
	int ch;

	luaZ_init();

	while (-1 != (ch = getopt_long(argc, argv, short_opts(long_opts), long_opts, 0))) {
		switch (ch) {
		case 'c':
			if (0 != luaL_dofile(opts.L, optarg))
				panic("%s", lua_tostring(opts.L, -1));

			break;
		case 'd':
			lua_getglobal(opts.L, "proc");
			lua_pushboolean(opts.L, 1);
			lua_setfield(opts.L, -2, "daemon");

			lua_pop(opts.L, 1);

			break;
		case 'p':
			lua_getglobal(opts.L, "proc");
			lua_pushstring(opts.L, optarg);
			lua_setfield(opts.L, -2, "pid-file");

			lua_pop(opts.L, 1);

			break;
		case 'w':
			lua_getglobal(opts.L, "proc");
			lua_pushboolean(opts.L, 1);
			lua_setfield(opts.L, -2, "watchdog");

			lua_pop(opts.L, 1);

			break;
		case 'e':
			lua_getglobal(opts.L, "log");
			lua_pushstring(opts.L, optarg);
			lua_setfield(opts.L, -2, "stderr");

			lua_pop(opts.L, 1);

			break;
		case 'u':
			do {
				char *d	= strchr(optarg, '.');

				lua_getglobal(opts.L, "proc");

				if (d) {
					lua_pushlstring(opts.L, optarg, d - optarg);
					lua_setfield(opts.L, -2, "uid");

					lua_pushstring(opts.L, d + 1);
					lua_setfield(opts.L, -2, "gid");
				} else {
					lua_pushstring(opts.L, optarg);
					lua_setfield(opts.L, -2, "uid");
				}

				lua_pop(opts.L, 1);
			} while (0);

			break;
		case 'l':
			luaZ_getglobal(opts.L, "proc", "listen");

			lua_pushinteger(opts.L, 1);
			lua_pushstring(opts.L, optarg);
			
			lua_settable(opts.L, -3);

			lua_pop(opts.L, 1);

			break;
		case 'q':
			luaZ_getglobal(opts.L, "db", "rep");

			lua_pushstring(opts.L, optarg);
			lua_setfield(opts.L, -2, "in-srv");

			lua_pop(opts.L, 1);

			break;
		case 'D':
			luaZ_getglobal(opts.L, "db", "env");

			lua_pushstring(opts.L, optarg);
			lua_setfield(opts.L, -2, "dir");

			lua_pop(opts.L, 1);

			break;
		case 'v': case -'v':
			lua_getglobal(opts.L, "log");
			lua_getfield(opts.L, -1, "level");

			if (optarg != 0) {
				lua_pushstring(opts.L, optarg);
			} else {
				int lvl;

				if (!lua_isnumber(opts.L, -1) ) {
					lua_pushinteger(opts.L, LOGGER_P_FATAL);
				} else switch ((lvl = lua_tointeger(opts.L, -1))) {
				case LOGGER_P_FATAL:
					lua_pushinteger(opts.L, LOGGER_P_CRITICAL);
					break;
				case LOGGER_P_CRITICAL:
					lua_pushinteger(opts.L, LOGGER_P_ERROR);
					break;
				default:
					lua_pushinteger(opts.L, ((LOGGER_P_DEBUG > lvl)? 5 : 1) + lvl);
				}
			}

			lua_replace(opts.L, -2);	/* replace old value */
			lua_setfield(opts.L, -2, "level");

			lua_pop(opts.L, 1);

			break;
		case 'o':
			if (0 != luaL_dostring(opts.L, optarg))
				panic("invalid option literal: %s", lua_tostring(opts.L, -1));

			lua_settop(opts.L, 0);

			break;
		case 'V':
			print_version(stdout); exit(EXIT_SUCCESS);
		case 'h':
			print_usage(stdout); exit(EXIT_SUCCESS);
		default:
			print_usage(stderr); exit(EXIT_FAILURE);
		}

		if (0 != lua_gettop(opts.L))
			panic("Lua stack out of order, top at %d but should be 0", lua_gettop(opts.L));
	} /* getopt_long() */

	if (optind < argc) {
		static const char *modes[] = { "start", "stop", "restart", 0 };

		lua_pushstring(opts.L, argv[optind++]);

		switch (luaL_checkoption(opts.L, -1, 0, modes)) {
		case 1: case 2:
			luaZ_getglobal(opts.L, "proc", "pid-file");

			if (lua_isnil(opts.L, -1))
				panic("must specify PID file to stop or restart");

			lua_pop(opts.L, 1);
		}

		lua_getglobal(opts.L, "main");
		lua_pushvalue(opts.L, -2);
		lua_setfield(opts.L, -2, "do");

		lua_pop(opts.L, 2);
	}

	if (0 != lua_gettop(opts.L))
		panic("Lua stack out of order, top at %d but should be 0", lua_gettop(opts.L));

	return /* void */;
} /* options_init() */


#undef luaZ_getfield

void luaZ_getfield(lua_State *L, int index, const char *name, ...) {
	const char *next;
	va_list ap;

	va_start(ap, name);

	lua_getfield(L, index, name);

	while (0 != (next = va_arg(ap, const char *)) && !lua_isnil(L, -1)) {
		if (lua_istable(L, -1)) {
			lua_getfield(L, -1, next);
		} else
			lua_pushnil(L);

		lua_replace(L, -2);
	}

	va_end(ap);

	return /* void */;
} /* luaZ_getfield() */

