/* ==========================================================================
 * pidfile.c - 'nix "pid file" implementation
 * --------------------------------------------------------------------------
 * Copyright (c) 2007  William Ahern
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ==========================================================================
 */
#include <stdio.h>	/* fopen(3) fclose(3) fscanf(3) setbuf(3) fileno(3) fprintf(3) */
#include <stdlib.h>	/* size_t malloc(3) free(3) */

#include <signal.h>	/* SIGTERM SIGKILL kill(2) */

#include <string.h>	/* strlen(3) */

#include <errno.h>	/* errno EWOULDBLOCK */

#include <sys/types.h>	/* pid_t mode_t */
#include <sys/stat.h>	/* umask(2) */
#include <sys/file.h>	/* LOCK_EX LOCK_UN LOCK_NB flock(2) */

#include <unistd.h>	/* getpid(2) ftruncate(2) */

#include "pidfile.h"


static int pidfile__nolock(struct pidfile *, void *);

static int pidfile__onlock(struct pidfile *, void *);

static int pidfile__unlock(struct pidfile *, void *);


const struct pidfile_options pidfile_defaults = {
	.mask	= ~0,

	.cb	= {
		.arg	= 0,
		.nolock	= &pidfile__nolock,
		.onlock	= &pidfile__onlock,
		.unlock	= &pidfile__unlock,
	},
};


struct pidfile {
	struct pidfile_options opts;

	char *path;

	FILE *fp;

	int have_lock;
}; /* struct pidfile */



static FILE *pidfile_fp_open(struct pidfile *p) {
	mode_t mask;
	FILE *fp;

	if (p->opts.mask != pidfile_defaults.mask)
		mask	= umask(p->opts.mask);

	fp	= fopen(p->path, "a+");

	if (p->opts.mask != pidfile_defaults.mask)
		(void)umask(mask);

	return fp;
} /* pidfile_fp_open() */


struct pidfile *pidfile_open(const char *path, const struct pidfile_options *opts) {
	static const struct pidfile p_initializer;
	struct pidfile *p	= 0;
	size_t plen;
	int syserr;

	if (0 == (p = malloc(sizeof *p)))
		return 0;

	*p	= p_initializer;
	p->opts	= *opts;

	if (0 == (p->path = malloc(1 + (plen = strlen(path)))))
		goto sysfail;

	(void)memcpy(p->path, path, plen + 1);

	if (0 == (p->fp = pidfile_fp_open(p)))
		goto sysfail;

	setbuf(p->fp, 0);

	return p;	
sysfail:
	syserr	= errno;

	free(p->path);
	free(p);

	errno	= syserr;

	return 0;
} /* pidfile_open() */


void pidfile_close(struct pidfile *p, int trunc) {
	if (p == 0)
		return /* void */;

	if (p->fp) {
		if (trunc)
			pidfile_setpid(p, PIDFILE_VOID_PID);

		fclose(p->fp), p->fp = 0;
	}

	if (p->opts.cb.nolock == &pidfile__nolock)
		free(p->opts.cb.arg), p->opts.cb.arg = 0;

	free(p->path), p->path = 0;

	free(p);

	return /* void */;
} /* pidfile_close() */


int pidfile_trylock(struct pidfile *p) {
	return (p->have_lock || 0 == flock(fileno(p->fp), LOCK_EX | LOCK_NB))? 0 : -1;
} /* pidfile_trylock() */


int pidfile_lock(struct pidfile *p) {
	int rval;

	while (0 != pidfile_trylock(p)) {
		if (0 != (rval = p->opts.cb.nolock(p, p->opts.cb.arg)))
			return rval;
	}

	p->have_lock	= 1;

	if (0 != p->opts.cb.onlock(p, p->opts.cb.arg))
		return -1;

	return 0;
} /* pidfile_lock() */


pid_t pidfile_getpid(struct pidfile *p) {
	int pid;

	rewind(p->fp);

	if (1 != fscanf(p->fp, "%d", &pid) || pid <= 0)
		return PIDFILE_VOID_PID;

	return pid;	
} /* pidfile_getpid() */


int pidfile_setpid(struct pidfile *p, pid_t pid) {
	if (0 != ftruncate(fileno(p->fp), 0))
		goto sysfail;

	if (pid != PIDFILE_VOID_PID
	&&  0 >= fprintf(p->fp, "%d\n", pid))
		goto sysfail;

	return 0;
sysfail:
	return -1;
} /* pidfile_setpid() */


struct pidfile__state {
	unsigned num_failures;

	pid_t last_pid;
}; /* struct pidfile__state */


static struct pidfile__state *pidfile__getstate(struct pidfile *p) {
	static const struct pidfile__state s_initializer = { .last_pid = PIDFILE_VOID_PID };
	struct pidfile__state *s = 0;

	if (p->opts.cb.arg != 0)
		return p->opts.cb.arg;

	if (0 == (s = malloc(sizeof *s)))
		return 0;

	*s	= s_initializer;

	return (p->opts.cb.arg = s);
} /* pidfile__getstate() */


static int pidfile__nolock(struct pidfile *p, void *arg) {
	struct pidfile__state *s;
	pid_t last_pid, this_pid;
	int syserr;

	if (errno != EWOULDBLOCK)
		goto sysfail;

	if (0 == (s = pidfile__getstate(p)))
		goto sysfail;

	last_pid	= s->last_pid;
	this_pid	= pidfile_getpid(p);
	s->last_pid	= this_pid;

	if (this_pid != PIDFILE_VOID_PID) {
		if (this_pid != last_pid)
			s->num_failures	= 0;

		switch (s->num_failures) {
		case 0:
			if (0 != kill(this_pid, SIGTERM))
				goto sysfail;

			break;
		case 3:
			(void)kill(this_pid, SIGKILL);

			break;
		case 5:
			goto sysfail;
		}
	}

	(void)sleep(!!s->num_failures++);

	return 0;
sysfail:
	syserr	= errno;

	free(p->opts.cb.arg), p->opts.cb.arg = 0;

	errno	= syserr;

	return -1;
} /* pidfile__nolock() */


static int pidfile__onlock(struct pidfile *p, void *arg) {
	return pidfile_setpid(p, PIDFILE_VOID_PID);
} /* pidfile__onlock() */


static int pidfile__unlock(struct pidfile *p, void *arg) {
	free(p->opts.cb.arg), p->opts.cb.arg = 0;

	return pidfile_setpid(p, PIDFILE_VOID_PID);
} /* pidfile__unlock() */


