/* ==========================================================================
 * zmbuf.c - Zoned - BIND DLZ Replicator
 * --------------------------------------------------------------------------
 * Copyright (C) 2007  William Ahern
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
 * ==========================================================================
 */
#include <stdlib.h>		/* malloc(3) calloc(3) free(3) */

#include <time.h>		/* time_t struct tm gmtime_r(3) */

#include <errno.h>		/* errno */

#include <sys/uio.h>		/* struct iovec */

#include "ZonedMessage.h"

#include "syserr.h"
#include "zmbuf.h"


asn_dec_rval_t zmsg_decode_ber(struct ZonedMessage **dst, const void *src, size_t slen) {
	asn_dec_rval_t rval;

	*dst	= 0;

	rval	= ber_decode(0, &asn_DEF_ZonedMessage, (void **)dst, src, slen);

	if (rval.code != RC_OK) {
		syserr_push(errno);

		zmsg_free(*dst), *dst = 0;

		syserr_pop();
	}

	return rval;
} /* zmsg_decode_ber() */


asn_dec_rval_t zmsg_decode_xer(struct ZonedMessage **dst, const void *src, size_t slen) {
	asn_dec_rval_t rval;

	*dst	= 0;

	rval	= xer_decode(0, &asn_DEF_ZonedMessage, (void **)dst, src, slen);

	if (rval.code != RC_OK) {
		syserr_push(errno);

		zmsg_free(*dst), *dst = 0;

		syserr_pop();
	}

	return rval;
} /* zmsg_decode_xer() */


static int zmsg_append(const void *bytes, size_t count, void *arg) {
	struct iovec *buf	= arg;
	unsigned char *p;

	errno	= 0;

	if (0 == (p = realloc(buf->iov_base, buf->iov_len + count)))
		return -1;

	errno	= 0;

	(void)memcpy(p + buf->iov_len, bytes, count);

	buf->iov_base	= p;
	buf->iov_len	+= count;

	return 0;	
} /* zmsg_append() */


asn_enc_rval_t zmsg_encode_der(void **dst, size_t *dlen, struct ZonedMessage *src) {
	struct iovec buf	= { *dst, 0 };
	asn_enc_rval_t rval;

	rval	= der_encode(&asn_DEF_ZonedMessage, src, &zmsg_append, &buf);

	if (rval.encoded == -1) {
		free(buf.iov_base);

		syserr_push(errno);

		buf.iov_base	= 0;
		buf.iov_len	= 0;

		syserr_pop();
	}

	*dst	= buf.iov_base;
	*dlen	= buf.iov_len;

	return rval;
} /* zmsg_encode_der() */


asn_enc_rval_t zmsg_encode_xer(void **dst, size_t *dlen, struct ZonedMessage *src) {
	struct iovec buf	= { *dst, 0 };
	asn_enc_rval_t rval;

	rval	= xer_encode(&asn_DEF_ZonedMessage, src, XER_F_BASIC, &zmsg_append, &buf);

	if (rval.encoded == -1) {
		free(buf.iov_base);

		syserr_push(errno);

		buf.iov_base	= 0;
		buf.iov_len	= 0;

		syserr_pop();
	}

	*dst	= buf.iov_base;
	*dlen	= buf.iov_len;

	return rval;
} /* zmsg_encode_xer() */


void zmsg_free(struct ZonedMessage *msg) {
	ASN_STRUCT_FREE(asn_DEF_ZonedMessage, msg);

	return /* void */;
} /* zmsg_free() */


struct ZonedMessage *zmsg_alloc(void) {
	static const struct ZonedMessage msg_initializer;
	struct ZonedMessage *msg;

	if (0 == (msg = malloc(sizeof *msg)))
		return 0;

	*msg	= msg_initializer;

	return msg;
} /* zmsg_alloc() */


struct ZonedMessage *zmsg_new(enum ZonedMessage_PR t) {
	struct ZonedMessage *msg;

	if (0 == (msg = zmsg_alloc()))
		return 0;
	
	msg->present	= t;

	return msg;
} /* zmsg_new() */


void zmbuf_free(struct zmbuf *mb) {
	if (mb == 0)
		return /* void */;

	LIST_REMOVE(mb, ref.le);

	if (LIST_EMPTY(mb->ref.refs)) {
		zmsg_free(mb->msg);

		free(mb);
	}

	return /* void */;
} /* zmbuf_free() */


struct zmbuf *zmbuf_alloc(void) {
	static const struct zmbuf mb_initializer;
	struct zmbuf *mb;

	if (0 == (mb = malloc(sizeof *mb)))
		return 0;

	*mb		= mb_initializer;
	mb->ref.refs	= &mb->ref.lh;

	LIST_INSERT_HEAD(mb->ref.refs, mb, ref.le);

	return mb;
} /* zmbuf_alloc() */


struct zmbuf *zmbuf_new(enum ZonedMessage_PR t) {
	struct zmbuf *mb;

	if (0 == (mb = zmbuf_alloc()))
		goto sysfail;

	if (0 == (mb->msg = zmsg_new(t)))
		goto sysfail;

	return mb;
sysfail:
	syserr_push(errno);

	zmbuf_free(mb);

	syserr_pop();

	return 0;
} /* zmbuf_new() */


struct zmbuf *zmbuf_dup(struct zmbuf *mb0) {
	struct zmbuf *mb;

	if (0 == (mb = zmbuf_alloc()))
		return 0;

	mb->ref.refs	= mb0->ref.refs;

	LIST_INSERT_HEAD(mb->ref.refs, mb, ref.le);

	mb->msg		= mb0->msg;

	return mb;
} /* zmbuf_dup() */


struct zmbuf *zmbuf_helo(const char *host, unsigned short port) {
	struct zmbuf *mb;

	if (0 == (mb = zmbuf_new(ZonedMessage_PR_helo)))
		goto sysfail;

	if (0 != OCTET_STRING_fromString(&mb->msg->helo.host.name, (char *)host))
		goto sysfail;

	mb->msg->helo.host.port	= port;

	return mb;
sysfail:
	syserr_push(errno);

	zmbuf_free(mb);

	syserr_pop();

	return 0;	
} /* zmbuf_helo() */


struct zmbuf *zmbuf_ping(time_t now) {
	struct zmbuf *mb;
	struct tm tmb;

	if (0 == (mb = zmbuf_new(ZonedMessage_PR_ping)))
		goto sysfail;

	if (0 == asn_time2GT(&mb->msg->ping.time, gmtime_r(&now, &tmb), 1))
		goto sysfail;

	return mb;
sysfail:
	syserr_push(errno);

	zmbuf_free(mb);

	syserr_pop();

	return 0;	
} /* zmbuf_ping() */


struct zmbuf *zmbuf_data(const DBT *control, const DBT *rec, const DB_LSN *lsnp) {
	struct zmbuf *mb;

	if (0 == (mb = zmbuf_new(ZonedMessage_PR_data)))
		goto sysfail;

	if (0 != OCTET_STRING_fromBuf(&mb->msg->data.control.data, control->data, control->size))
		goto sysfail;

	if (0 != OCTET_STRING_fromBuf(&mb->msg->data.rec.data, rec->data, rec->size))
		goto sysfail;

	mb->msg->data.lsnp.file		= lsnp->file;
	mb->msg->data.lsnp.offset	= lsnp->offset;

	return mb;
sysfail:
	syserr_push(errno);

	zmbuf_free(mb);

	syserr_pop();

	return 0;
} /* zmbuf_data() */


struct zmbuf *zmbuf_edit(const char *zone, enum ZonedEditMode mode) {
	struct zmbuf *mb;

	if (0 == (mb = zmbuf_new(ZonedMessage_PR_edit)))
		goto sysfail;

	if (0 != OCTET_STRING_fromString(&mb->msg->edit.zone, zone))
		goto sysfail;

	if (0 != asn_long2INTEGER(&mb->msg->edit.mode, mode))
		goto sysfail;

	return mb;
sysfail:
	syserr_push(errno);

	zmbuf_free(mb);

	syserr_pop();

	return 0;
} /* zmbuf_edit() */


int zmbuf_edit_any(struct zmbuf *mb, const char *host, unsigned ttl, const char *rtype, const char *rdata) {
	struct ZonedRR *rr;

	if (0 == (rr = calloc(sizeof *rr, 1)))	
		goto sysfail;

	rr->present	= ZonedRR_PR_any;

	if (0 != OCTET_STRING_fromString(&rr->any.host, host))
		goto sysfail;

	rr->any.ttl	= ttl;

	if (0 != OCTET_STRING_fromString(&rr->any.rtype, rtype))
		goto sysfail;

	if (0 != OCTET_STRING_fromString(&rr->any.rdata, rdata))
		goto sysfail;

	if (0 != ASN_SET_ADD(&mb->msg->edit.rrset, rr))
		goto sysfail;

	return 0;
sysfail:
	syserr_push(errno);

	free(rr), rr = 0;

	syserr_pop();

	return -1;
} /* zmbuf_edit_any() */


int zmbuf_edit_a(struct zmbuf *mb, const char *host, unsigned ttl, const char *address) {
	struct ZonedRR *rr;

	if (0 == (rr = calloc(sizeof *rr, 1)))	
		goto sysfail;

	rr->present	= ZonedRR_PR_a;

	if (0 != OCTET_STRING_fromString(&rr->a.host, host))
		goto sysfail;

	rr->a.ttl	= ttl;

	if (0 != OCTET_STRING_fromString(&rr->a.address, address))
		goto sysfail;

	if (0 != ASN_SET_ADD(&mb->msg->edit.rrset, rr))
		goto sysfail;

	return 0;
sysfail:
	syserr_push(errno);

	free(rr), rr = 0;

	syserr_pop();

	return -1;
} /* zmbuf_edit_a() */


int zmbuf_edit_srv(struct zmbuf *mb, const char *host, unsigned ttl, unsigned priority, unsigned weight, unsigned short port, const char *target) {
	struct ZonedRR *rr;

	if (0 == (rr = calloc(sizeof *rr, 1)))	
		goto sysfail;

	rr->present	= ZonedRR_PR_srv;

	if (0 != OCTET_STRING_fromString(&rr->srv.host, host))
		goto sysfail;

	rr->srv.ttl		= ttl;
	rr->srv.priority	= priority;
	rr->srv.weight		= weight;
	rr->srv.port		= port;

	if (0 != OCTET_STRING_fromString(&rr->srv.target, target))
		goto sysfail;

	if (0 != ASN_SET_ADD(&mb->msg->edit.rrset, rr))
		goto sysfail;

	return 0;
sysfail:
	syserr_push(errno);

	free(rr), rr = 0;

	syserr_pop();

	return -1;
} /* zmbuf_edit_srv() */


