/* ==========================================================================
 * Zoned - BIND DLZ Replicator
 * --------------------------------------------------------------------------
 * Copyright (C) 2007  William Ahern
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * ==========================================================================
 */
#include <stdlib.h>		/* malloc(3) free(3) */

#include <stdio.h>		/* snprintf(3) */

#include <limits.h>		/* INT_MAX */

#include <time.h>		/* time(2) */

#include <signal.h>		/* SIGUSR1 */

#include <string.h>		/* strcasecmp(3) strerror(3) memmove(3) */

#include <errno.h>		/* errno */

#include <assert.h>

#include <sys/types.h>		/* socklen_t */
#include <sys/socket.h>		/* struct sockaddr_storage struct sockaddr */
#include <sys/queue.h>		/* LIST SIMPLEQ CIRCLEQ */
#include <sys/tree.h>		/* RB */

#include <netinet/in.h>		/* AF_INET AF_INET6 struct sockaddr_in struct sockaddr_in6 */

#include <arpa/inet.h>		/* htons(3) */

#include <netdb.h>		/* NI_MAXHOST NI_MAXSERV getnameinfo(3) */

#include <unistd.h>		/* gethostname(3) */

#include <pthread.h>		/* pthread_mutex_t pthread_kill(2) */

#include <db.h>			/* all kinds of scariness */

#include <evnet/lookup.h>	/* LOOKUP_IN_A */
#include <evnet/socket.h>	/* struct socket */
#include <evnet/bufio.h>	/* struct bufio */

#include <sys/param.h>		/* arraylen */

#include "ZonedMessage.h"

#include "syserr.h"
#include "options.h"
#include "log.h"
#include "zmbuf.h"
#include "bdb.h"
#include "edit.h"
#include "znode.h"


#ifndef ZONED_PING_INTERVAL
#define ZONED_PING_INTERVAL	20
#endif


static pthread_t control_loop;


#define znode_async_lock(z)	pthread_mutex_lock(&znodes.mutex)
#define znode_async_unlock(z)	pthread_mutex_unlock(&znodes.mutex)

#define znode_isalive(z)	 ((z)->closed == 0)

struct znode {
	pthread_mutex_t mutex;

	int refc;

	int closed;

	char label[256];

	struct {
		char name[256];
		unsigned short port;
	} host;

	struct {
		int pending;
		struct event evb;
	} timer;

	struct socket *so;
	struct bufio *iob;

	struct {
		int pending;

		unsigned char *buf;
		unsigned char *pos;
		unsigned char *end;

		SIMPLEQ_HEAD(, zmbuf) mbufs;

		SIMPLEQ_HEAD(, zmbuf) tbufs;	/* thread pass queue */

		LIST_ENTRY(znode) le;

		CIRCLEQ_ENTRY(znode) cqe;
	} rd, wr;

	struct {
		int envid;
	} db;

	struct {
		RB_ENTRY(znode) byenvid, byhost;
	} rbe;
}; /* struct znode */


static struct znode zself_ = {
	.db	= { .envid = BDB_SELF_EID },
};

struct znode *zself	= &zself_;


static struct {
	pthread_mutex_t mutex;

	RB_HEAD(znodes_byenvid, znode) byenvid;
	RB_HEAD(znodes_byhost, znode) byhost;

	int noenvid, nohost;

	struct {
		LIST_HEAD(, znode) ready, none;
	} rd, wr;

	CIRCLEQ_HEAD(, znode) dirty;	/* touched async */

	struct event sigeb;
} znodes = {
	.mutex		= PTHREAD_MUTEX_INITIALIZER,
	.byenvid	= RB_INITIALIZER(&znodes.byenvid),
	.byhost		= RB_INITIALIZER(&znodes.byhost),
	.dirty		= CIRCLEQ_HEAD_INITIALIZER(znodes.dirty),
};


static inline int znode_cmp_byenvid(struct znode *, struct znode *);

RB_GENERATE(znodes_byenvid, znode, rbe.byenvid, znode_cmp_byenvid);

static inline int znode_cmp_byenvid(struct znode *a, struct znode *b) {
	return a->db.envid - b->db.envid;
} /* znode_cmp_byenvid() */


static inline int znode_cmp_byhost(struct znode *, struct znode *);

RB_GENERATE(znodes_byhost, znode, rbe.byhost, znode_cmp_byhost);

static inline int znode_cmp_byhost(struct znode *a, struct znode *b) {
	int cmp;

	if (0 != (cmp = strcasecmp(a->host.name, b->host.name)))
		return cmp;
	else
		return (int)a->host.port - (int)b->host.port;
} /* znode_cmp_byhost() */


int znode_count_byhost(void) {
	return znodes.nohost;
} /* znode_count_byhost() */


int znode_count_byenvid(void) {
	return znodes.noenvid;
} /* znode_count_byenvid() */


struct znode *znode_find_byhost(const char *host, unsigned short port) {
	struct znode k;

	assert(0 < snprintf(k.host.name, sizeof k.host.name, "%s", host));

	k.host.port	= port;

	return RB_FIND(znodes_byhost, &znodes.byhost, &k);
} /* znode_find_byhost() */


struct znode *znode_find_byenvid(int envid) {
	struct znode k;

	k.db.envid	= envid;

	return RB_FIND(znodes_byenvid, &znodes.byenvid, &k);
} /* znode_find_byenvid() */


static void znode_rokay(struct znode *z, int ok) {
	LIST_REMOVE(z, rd.le);

	if (ok)
		LIST_INSERT_HEAD(&znodes.rd.ready, z, rd.le);
	else
		LIST_INSERT_HEAD(&znodes.rd.none, z, rd.le);

	return /* void */;
} /* znode_rokay() */


static void znode_wokay(struct znode *z, int ok) {
	LIST_REMOVE(z, wr.le);

	if (ok)
		LIST_INSERT_HEAD(&znodes.wr.ready, z, wr.le);
	else
		LIST_INSERT_HEAD(&znodes.wr.none, z, wr.le);

	return /* void */;
} /* znode_wokay() */


void znode_send(struct znode *z, struct zmbuf *mb) {
	SIMPLEQ_INSERT_TAIL(&z->wr.mbufs, mb, sqe);

	znode_wokay(z, 1);

	return /* void */;
} /* znode_send() */


static void znode_runq(void);

static void zmbuf_sweep(int fd, short events, void *arg) {
	struct znode *z;
	struct zmbuf *mb;

	debug2("sweeping dirty DATA");

	znode_async_lock();

	CIRCLEQ_FOREACH(z, &znodes.dirty, wr.cqe) {
		if (SIMPLEQ_EMPTY(&z->wr.tbufs))
			break;

		while (SIMPLEQ_END(&z->wr.tbufs) != (mb = SIMPLEQ_FIRST(&z->wr.tbufs))) {
			SIMPLEQ_REMOVE_HEAD(&z->wr.tbufs, sqe);
			znode_send(z, mb);
		}
	}
	
	znode_async_unlock();

	znode_runq();

	return /* void */;
} /* zmbuf_sweep() */


int zmbuf_send(struct zmbuf *mb, int envid, uint32_t flags) {
	int rval	= 0;
	struct znode *z;

	znode_async_lock();

	if (envid == DB_EID_BROADCAST) {
		debug1("sending broadcast message");

		RB_FOREACH(z, znodes_byhost, &znodes.byhost) {
			if (z == zself)
				continue;

			SIMPLEQ_INSERT_TAIL(&z->wr.tbufs, mb, sqe);

			CIRCLEQ_REMOVE(&znodes.dirty, z, wr.cqe);
			CIRCLEQ_INSERT_HEAD(&znodes.dirty, z, wr.cqe);

			if (0 == (mb = zmbuf_dup(mb))) {
				rval	= -1;

				goto epilog;
			}
		}

		zmbuf_free(mb), mb = 0;
	} else if (0 != (z = znode_find_byenvid(envid))) {
		assert(z != zself);

		debug1("sending message to %s", z->label);

		SIMPLEQ_INSERT_TAIL(&z->wr.tbufs, mb, sqe);

		CIRCLEQ_REMOVE(&znodes.dirty, z, wr.cqe);
		CIRCLEQ_INSERT_HEAD(&znodes.dirty, z, wr.cqe);
	} else {
		debug0("unknown node: %d", envid);

		rval	= -1;
	}
epilog:
	znode_async_unlock();

	debug2("raise(SIGUSR1)");

	pthread_kill(control_loop, SIGUSR1);

	return rval;
} /* zmbuf_send() */


static void znode_catch_recv(struct znode *z, size_t len) {
	struct ZonedMessage *msg	= 0;
	asn_dec_rval_t rval;

	z->rd.pending	= 0;

	z->rd.pos	+= len;

next:
	rval	= zmsg_decode(&msg, z->rd.buf, z->rd.pos - z->rd.buf);

	if (rval.code == RC_OK) {
		struct zmbuf *mb;

		(void)memmove(z->rd.buf, &z->rd.buf[rval.consumed], z->rd.pos - &z->rd.buf[rval.consumed]);
		z->rd.pos	-= rval.consumed;

		if (0 != (mb = zmbuf_alloc())) {
			mb->msg	= msg;

			SIMPLEQ_INSERT_TAIL(&z->rd.mbufs, mb, sqe);

			znode_rokay(z, 1);

			goto next;
		} else {
			error("unable to allocate znode mbuf: %s", strerror(errno));

			goto anyfail;
		}
	} else if (rval.code == RC_FAIL) {
		warn("(%s) corrupt message", z->label);

		goto anyfail;
	}

	return /* void */;
anyfail:
	znode_close(z);

	return /* void */;
} /* znode_catch_recv() */


static void znode_catch_send(struct znode *z, size_t len) {
	struct zmbuf *mb;

	z->wr.pending	= 0;

	if (len != (z->wr.pos - z->wr.buf)) {
		notice("(%s) hangup", z->label);

		znode_close(z);

		return /* void */;
	}

	mb	= SIMPLEQ_FIRST(&z->wr.mbufs);

	SIMPLEQ_REMOVE_HEAD(&z->wr.mbufs, sqe);

	znode_wokay(z, SIMPLEQ_EMPTY(&z->wr.mbufs));

	zmbuf_free(mb);

	return /* void */;	
} /* znode_catch_send() */


static void znode_run(struct znode *);

static void znode_catch_helo(struct znode *z, struct zmbuf *mb) {
	enum socket_mode mode;

	debug("(%s) received HELO: [%.*s]:%ld", z->label, (int)mb->msg->helo.host.name.size, (char *)mb->msg->helo.host.name.buf, mb->msg->helo.host.port);

	socket_getoption(z->so, SOCKET_SO_MODE, &mode);

	if (*z->host.name == '\0') {
		if (mb->msg->helo.host.name.size >= sizeof z->host.name || 0 >= mb->msg->helo.host.name.size) {
			warn("(%s) invalid host name (%d bytes)", z->label, (int)mb->msg->helo.host.name.size);

			goto anyfail;
		}

		(void)memcpy(z->host.name, mb->msg->helo.host.name.buf, mb->msg->helo.host.name.size);
		z->host.name[mb->msg->helo.host.name.size]	= '\0';

		if (mb->msg->helo.host.port <= 0 || mb->msg->helo.host.port > 65535) {
			warn("(%s) invalid port (%ld)", z->label, mb->msg->helo.host.port);

			goto anyfail;
		}

		z->host.port	= mb->msg->helo.host.port;

		if (0 == strcasecmp(z->host.name, zself->host.name)
		&&  z->host.port == zself->host.port) {
			debug("(%s) talking to self", z->label);

			goto anyfail;
		}

		znode_async_lock();

		if (0 != RB_INSERT(znodes_byhost, &znodes.byhost, z)) {
			notice("(%s) already connected ([%s]:%hu", z->label, z->host.name, z->host.port);

			znode_async_unlock();

			goto anyfail;
		}

		znodes.nohost++;

		znode_async_unlock();

		debug1("inserted [%s]:%hu (#%d)", z->host.name, z->host.port, znodes.nohost);
	} else if (strlen(z->host.name) != mb->msg->helo.host.name.size || 0 != strncasecmp((char *)mb->msg->helo.host.name.buf, z->host.name, mb->msg->helo.host.name.size) || mb->msg->helo.host.port != z->host.port) {
		warn("(%s) helo mismatch ([%s]:%hu / [%.*s]:%ld)", z->label, z->host.name, z->host.port, (int)mb->msg->helo.host.name.size, (char *)mb->msg->helo.host.name.buf, mb->msg->helo.host.port);

		goto anyfail;
	}

	socket_getoption(z->so, SOCKET_SO_MODE, &mode);

	if (mode == SOCKET_MODE_SERVER) {
		struct zmbuf *mb;

		if (0 == (mb = zmbuf_helo(zself->host.name, zself->host.port))) {
			error("(%s) unable to allocate HELO message: %s", z->label, strerror(errno));

			goto anyfail;
		}

		znode_send(z, mb);
	}

	bdb_wakeup(BDB_EV_NODE_JOIN);

	goto epilog;
anyfail:
	znode_close(z), z = 0;
epilog:
	zmbuf_free(mb);

	return /* void */;	
} /* znode_catch_helo() */


static void znode_catch_ping(struct znode *z, struct zmbuf *mb) {
	debug("(%s) received PING", z->label);

	zmbuf_free(mb);

	return /* void */;
} /* znode_catch_ping() */


static void znode_catch_data(struct znode *z, struct zmbuf *mb) {
	debug("(%s) received DATA", z->label);

	zmbuf_recv(mb, z->db.envid);

	return /* void */;
} /* znode_catch_data() */


static void znode_catch_edit(struct znode *z, struct zmbuf *mb) {
	debug("(%s) received EDIT", z->label);

	edit_apply(mb);

	return /* void */;
} /* znode_catch_edit() */


static void znode_dispatch(struct znode *z, struct zmbuf *mb) {
	static const typeof(&znode_dispatch) dispatch[]	= {
		[ZonedMessage_PR_helo]	= &znode_catch_helo,
		[ZonedMessage_PR_ping]	= &znode_catch_ping,
		[ZonedMessage_PR_data]	= &znode_catch_data,
		[ZonedMessage_PR_edit]	= &znode_catch_edit,
	};
	enum ZonedMessage_PR type	= mb->msg->present;

	if (type >= 0 && type < arraylen(dispatch) && 0 != dispatch[type]) {
		dispatch[type](z, mb);
	} else {
		warn("(%s) unknown message: %d", z->label, mb->msg->present);

		zmbuf_free(mb);
	}

	return /* void */;	
} /* znode_dispatch() */


static void znode_wake(int fd, short events, void *arg) {
	struct znode *z	= arg;
	struct zmbuf *mb;

	z->timer.pending	= 0;

	if (0 == (mb = zmbuf_ping(time(0)))) {
		error("unable to allocate PING message: %s", strerror(errno));

		znode_close(z);

		return /* void */;
	}

	znode_send(z, mb);

	znode_run(z);

	return /* void */;
} /* znode_wake() */


static void znode_catch(struct bufio *iob, void *p, size_t len, enum bufio_errno iob_err, void *arg) {
	struct znode *z	= arg;

	if (iob_err != 0) {
		if (iob_err == BUFIO_EEOF)
			notice("(%s) hangup", z->label);
		else
			notice("(%s) I/O error: %s", z->label, (iob_err == BUFIO_ESYSTEM)? strerror(errno) : bufio_errlist[iob_err]);

		znode_close(z);

		return /* void */;
	} else if (len == 0) {
		critical("(%s) zero length I/O request", z->label);

		znode_close(z);

		return /* void */;
	}

	znode_refinc(z);

	if (z->wr.buf == p)
		znode_catch_send(z, len);
	else
		znode_catch_recv(z, len);

	znode_run(z);

	znode_refdec(z);

	return /* void */;
} /* znode_catch() */


static void znode_run(struct znode *z) {
	struct zmbuf *mb;

	if (!znode_isalive(z))
		return /* void */;

	znode_refinc(z);

again:
	if (!z->wr.pending && !SIMPLEQ_EMPTY(&z->wr.mbufs)) {
		struct timeval timeout	= { 30, 0 };
		asn_enc_rval_t rval;
		size_t len;

		len		= z->wr.end - z->wr.buf;
		rval		= zmsg_encode((void **)&z->wr.buf, &len, SIMPLEQ_FIRST(&z->wr.mbufs)->msg);
		z->wr.pos	= z->wr.end
				= z->wr.buf + len;

		if (rval.encoded == -1) {
			if (rval.failed_type != 0 && rval.failed_type->name != 0)
				warn("unable to DER encode: %s", rval.failed_type->name);
			else
				warn("unable to DER encode");

			znode_close(z);

			goto epilog;
		}

		z->wr.pending	= 1;

		bufio_writen(z->iob, z->wr.buf, z->wr.pos - z->wr.buf, &znode_catch, z, &timeout);

		if (!znode_isalive(z))
			goto epilog;
	}

	if (!z->rd.pending) {
		struct timeval timeout = { 2 * ZONED_PING_INTERVAL, 0 };
		unsigned char *tmp;

		if (z->rd.end - z->rd.pos < 512) {
			if (0 == (tmp = realloc(z->rd.buf, (z->rd.pos - z->rd.buf) + 512))) {
				error("realloc: %s", strerror(errno));

				znode_close(z);

				goto epilog;
			}

			z->rd.pos	= tmp + (z->rd.pos - z->rd.buf);
			z->rd.end	= z->rd.pos + 512;
			z->rd.buf	= tmp;
		}

		z->rd.pending	= 1;

		bufio_read(z->iob, z->rd.pos, z->rd.end - z->rd.pos, &znode_catch, z, &timeout);

		if (!znode_isalive(z))
			goto epilog;
	}

	while (znode_isalive(z) && SIMPLEQ_END(&z->rd.mbufs) != (mb = SIMPLEQ_FIRST(&z->rd.mbufs))) {
		SIMPLEQ_REMOVE_HEAD(&z->rd.mbufs, sqe);

		znode_dispatch(z, mb);
	}

	znode_rokay(z, 0);

	if (znode_isalive(z) && !z->wr.pending && !SIMPLEQ_EMPTY(&z->wr.mbufs))
		goto again;

	if (znode_isalive(z) && !z->timer.pending) {
		struct timeval timeout = { ZONED_PING_INTERVAL, 0 };

		timeout_set(&z->timer.evb, &znode_wake, z);

		if (0 != timeout_add(&z->timer.evb, &timeout)) {
			error("timeout_add: %s", strerror(errno));

			znode_close(z);

			goto epilog;
		}

		z->timer.pending	= 1;
	}

epilog:
	znode_refdec(z);

	return /* void */;
} /* znode_run() */


static void znode_runq(void) {
	struct znode *z, *n;

	for (z = LIST_FIRST(&znodes.wr.ready); z != LIST_END(&znodes.wr.ready); z = n) {
		n	= LIST_NEXT(z, wr.le);

		znode_run(z);
	}

	for (z = LIST_FIRST(&znodes.rd.ready); z != LIST_END(&znodes.rd.ready); z = n) {
		n	= LIST_NEXT(z, rd.le);

		znode_run(z);
	}

	return /* void  */;
} /* znode_runq() */


void znode_start(struct znode *z) {
	enum socket_mode mode;
	struct zmbuf *mb;

	socket_getoption(z->so, SOCKET_SO_MODE, &mode);

	if (mode == SOCKET_MODE_CLIENT) {
		if (0 == (mb = zmbuf_helo(zself->host.name, zself->host.port)))
			goto sysfail;

		znode_send(z, mb);
	}

	notice("(%s) starting znode", z->label);

	znode_run(z);

	return /* void */;
sysfail:
	warn("(%s) failed to start znode: %s", z->label, strerror(errno));

	znode_close(z);

	return /* void */;
} /* znode_start() */


static void znode__close(struct znode *z) {
	int was_host	= 0;
	struct zmbuf *mb;

	if (*z->label)
		notice("(%s) closing down znode", z->label);

	znode_async_lock();

	if (z == RB_FIND(znodes_byenvid, &znodes.byenvid, z))  {
		RB_REMOVE(znodes_byenvid, &znodes.byenvid, z);
		znodes.noenvid--;
	}

	if (z == RB_FIND(znodes_byhost, &znodes.byhost, z)) {
		RB_REMOVE(znodes_byhost, &znodes.byhost, z);
		znodes.nohost--;

		was_host	= 1;
	}

	CIRCLEQ_REMOVE(&znodes.dirty, z, wr.cqe);

	znode_async_unlock();

	LIST_REMOVE(z, rd.le);
	LIST_REMOVE(z, wr.le);

	free(z->rd.buf), z->rd.buf = z->rd.pos = z->rd.end = 0;

	while (SIMPLEQ_END(&z->rd.mbufs) != (mb = SIMPLEQ_FIRST(&z->rd.mbufs))) {
		SIMPLEQ_REMOVE_HEAD(&z->rd.mbufs, sqe);

		zmbuf_free(mb);
	}

	free(z->wr.buf), z->wr.buf = z->wr.pos = z->wr.end = 0;

	while (SIMPLEQ_END(&z->wr.mbufs) != (mb = SIMPLEQ_FIRST(&z->wr.mbufs))) {
		SIMPLEQ_REMOVE_HEAD(&z->wr.mbufs, sqe);

		zmbuf_free(mb);
	}

	while (SIMPLEQ_END(&z->wr.tbufs) != (mb = SIMPLEQ_FIRST(&z->wr.tbufs))) {
		SIMPLEQ_REMOVE_HEAD(&z->wr.tbufs, sqe);

		zmbuf_free(mb);
	}

	if (z->timer.pending)
		timeout_del(&z->timer.evb);

	bufio_close(z->iob), z->iob = 0;
	socket_close(z->so), z->so = 0;

	free(z), z = 0;

	if (was_host)
		bdb_wakeup(BDB_EV_NODE_GONE);

	return /* void */;
} /* znode__close() */


void znode_close(struct znode *z) {
	if (z == 0)
		return /* void */;

	z->closed	= 1;

	znode_refdec(z);

	return /* void */;
} /* znode_close() */


static int znode_setlabel(struct znode *z, struct socket *so) {
	struct sockaddr_storage ss;
	socklen_t salen;
	char host[NI_MAXHOST];
	char serv[NI_MAXSERV];
	const char *dname;
	enum socket_errno s_err;
	int g_err;

	if (0 != (s_err = socket_getpeername(so, (struct sockaddr *)&ss, (salen = sizeof ss, &salen), &dname))) {
		critical("unable to get socket peer name: %s", socket_strerror(s_err));

		return -1;
	}

	if (0 != (g_err = getnameinfo((struct sockaddr *)&ss, salen, host, sizeof host, serv, sizeof serv, NI_NUMERICHOST|NI_NUMERICSERV))) {
		critical("unable to get socket peer name: %s", gai_strerror(g_err));

		return -1;
	}

	if (dname != 0)
		assert(0 < snprintf(host, sizeof host, "%s", dname));

	assert(0 < snprintf(z->label, sizeof z->label, "[%s]:%s", host, serv));

	return 0;
} /* znode_setlabel() */


static int znode_envid_inc(int *envid)
	{ return (*envid == INT_MAX)? (*envid = BDB_SELF_EID + 1) : ++*envid; }

struct znode *znode_open(struct socket *s) {
	static const struct znode z_initializer	= { .mutex = PTHREAD_MUTEX_INITIALIZER };
	static int envidc	= BDB_SELF_EID;
	struct znode *z		= 0;
	enum bufio_errno err;

	if (0 == (z = malloc(sizeof *z)))
		return 0;

	*z		= z_initializer;

	SIMPLEQ_INIT(&z->rd.mbufs);
	SIMPLEQ_INIT(&z->wr.mbufs);
	SIMPLEQ_INIT(&z->wr.tbufs);

	LIST_INSERT_HEAD(&znodes.rd.none, z, rd.le);
	LIST_INSERT_HEAD(&znodes.wr.none, z, wr.le);

	if (0 != znode_setlabel(z, s))
		goto anyfail;

	znode_async_lock();

	CIRCLEQ_INSERT_TAIL(&znodes.dirty, z, wr.cqe);

	do {
		/* Technically, we could loop forever ;) */
		z->db.envid	= znode_envid_inc(&envidc);
	} while (0 != RB_INSERT(znodes_byenvid, &znodes.byenvid, z));

	znodes.noenvid++;

	znode_async_unlock();

	if (0 == (z->iob = bufio_open(&bufio_defaults, 0, &err)))
		goto sysfail;
	if (0 != bufio_set_sink(z->iob, socket_to_sink(s)))
		goto sysfail;
	if (0 != bufio_set_source(z->iob, socket_to_source(s)))
		goto sysfail;

	z->so	= s;

	znode_refinc(z);

	return z;
sysfail:
	error("unable to allocate node: %s", strerror(errno));
anyfail:
	syserr_push(errno);

	znode__close(z);

	syserr_pop();

	return 0;
} /* znode_open() */


void znode_refinc(struct znode *z) {
	++z->refc;

	return /* void */;
} /* znode_refinc() */


void znode_refdec(struct znode *z) {
	if (0 >= --z->refc)
		znode__close(z);

	return /* void */;
} /* znode_refdec() */


const char *znode_label(struct znode *z) {
	return z->label;
} /* znode_label() */


void znode_init(void) {
	struct sockaddr *sa	= 0;

	luaZ_getglobal(opts.L, "proc", "listen");

	lua_pushnil(opts.L);

	while (sa == 0 && 0 != lua_next(opts.L, -2)) {
		if (0 != (sa = luaL_checkudata(opts.L, -1, "sockaddr"))) {
			if (sa->sa_family != AF_INET && sa->sa_family != AF_INET6)
				sa	= 0;
		}

		lua_pop(opts.L, 1);
	}

	lua_settop(opts.L, 0);

	zself->db.envid	= BDB_SELF_EID;

	if (sa == 0 || 0 == (zself->host.port = htons(((struct sockaddr_in *)sa)->sin_port)))
		panic("unable to determine publishable port");

	if (0 != gethostname(zself->host.name, sizeof zself->host.name))
		panic("gethostname: %s", strerror(errno));

	RB_INSERT(znodes_byenvid, &znodes.byenvid, zself);
	znodes.noenvid++;

	RB_INSERT(znodes_byhost, &znodes.byhost, zself);
	znodes.nohost++;

	control_loop	= pthread_self();

	signal_set(&znodes.sigeb, SIGUSR1, &zmbuf_sweep, 0);

	if (0 != signal_add(&znodes.sigeb, 0))
		panic("signal_add: %s", strerror(errno));

	return /* void */;
} /* znode_init() */
